#!/usr/bin/env python
# ******************************************************************************
# Copyright 2023 Brainchip Holdings Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ******************************************************************************
import numpy as np
from onnxruntime.quantization.calibrate import TensorData

from .fixed_point import to_fixed_point
from .input_scale import input_scale_no_zp
from .weights import align_to


def downscale(output_range, i_scale, force_fp=False, bitwidth=8):
    """Calculates the scale that should be applied to an integer tensor
    with i_scale to project it to a desired bitwidth.

    The following set of operations must be applied to the tensor to project it
    into the output scale:

    >>> out_tensor = tensor * scale
    >>> out_tensor = out_tensor >> log2(shift)

    Args:
        output_range (tuple): the calibrate tensor range
        i_scale (np.ndarray): the input scale
        force_fp (bool, optional): whether to force output scale as a power-of-two.
            Defaults to False.
        bitwidth (int, optional): the desired output bitwidth. Defaults to 8.

    Returns:
        np.ndarray, np.ndarray, np.ndarray: the integer scale/shift and the new float scale
    """
    if force_fp:
        # The multi-input layers supported in akida (such as Add) do not include a scale-in
        # operation but only a shift-in. In consequence output must be downscaled as a fixed-point.
        return downscale_fp(output_range, i_scale, bitwidth=bitwidth)
    return downscale_qf(output_range, i_scale, bitwidth)


def downscale_qf(output_range, i_scale, bitwidth=8):
    # Consider all outputs to be 8-bits, otherwise the scale would be different.
    ocalib_scale = input_scale_no_zp(output_range)
    # Divide o_calib_scale by i_scale in the same axis to obtain output scale:
    # this will consider the input scale into account.
    o_scale = align_to(ocalib_scale, i_scale.ndim) / i_scale
    # Quantize o_scale to fit in scale + shift at 8 bit
    scale, shift = to_fixed_point(o_scale, bitwidth=bitwidth, signed=False)
    # Return shift value as a power of two
    s_out = np.array(2. ** shift, dtype=np.float32)
    return scale, s_out, np.array(ocalib_scale, "float64")


def downscale_fp(output_range, i_scale, bitwidth=8):
    assert isinstance(output_range, TensorData)
    # Dequantize inputs in integer domain (apply scale out), multiplying by the inverse scale
    scale, in_shift = to_fixed_point(1.0 / i_scale, bitwidth=bitwidth, signed=False)
    # Compute the required output shift to come out in 8bits
    output_max = np.maximum(np.abs(output_range.lowest), np.abs(output_range.highest))
    _, out_shift = to_fixed_point(output_max, bitwidth=bitwidth, signed=True, clamp=True)
    # Compute shift to go from in_shift to out_shift in the same axis
    # The shift can be positive (left-shift) or negative (rounded right-shift)
    shift = align_to(out_shift, i_scale.ndim) - in_shift
    # A positive shift exceeding the target bitwidth always leads to a saturation
    np.testing.assert_array_less(shift, bitwidth,
                                 f"Cannot rescale inputs to {bitwidth} as it will saturate.")
    # In ONNX output shift is done as division (against to akida: a left shift)
    shift = np.array(2. ** -shift, dtype=np.float32)
    # Finally, outputs will have a fractional scale
    o_scale = np.array(2.0 ** out_shift, "float64")
    return scale, shift, o_scale



#!/usr/bin/env python
# ******************************************************************************
# Copyright 2024 Brainchip Holdings Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ******************************************************************************
__all__ = ["make_fn_on_list", "remove_outliers", "compare", "eval_metrics"]

import numpy as np

from ..tools import SMAPE, Saturation


def make_fn_on_list(func):
    """Decorator that handles list arguments to a function.

    Call the function recursively for each iterable-element in the arguments.
    Stops when the arguments are no longer a list.

    Note auxiliary key-arguments are propagated for each call with their value.

    Args:
        func (callable): function to call.
    """
    def wrapper_fn(*args, **kwargs):
        types = (list, tuple)
        if any(isinstance(x, types) for x in args):
            assert all(isinstance(x, types) for x in args), "All args must be a list or tuple"
            assert all(len(args[0]) == len(x) for x in args), "All args must have the same length"
            return [wrapper_fn(*new_args, **kwargs) for new_args in zip(*args)]
        return func(*args, **kwargs)
    return wrapper_fn


def make_fn_on_dict(func):
    """Decorator that handles dictionaries arguments to a function.

    Call the function recursively for each iterable-element in the arguments.
    Stops when the arguments are no longer a dictionnary.

    Note auxiliary key-arguments are propagated for each call with their value.

    Args:
        func (callable): function to call.
    """
    def wrapper_fn(*args, **kwargs):
        if any(isinstance(x, dict) for x in args):
            assert all(isinstance(x, dict) for x in args), "All args must be a dict"
            keys = args[0].keys()
            if not all(k in keys for x in args for k in x):
                raise KeyError("All args must have the same keys")
            return {k: wrapper_fn(*tuple(x[k] for x in args), **kwargs) for k in keys}
        return func(*args, **kwargs)
    return wrapper_fn


def remove_outliers(values, mask, /, axis=None):
    """Remove outliers marked by mask

    Args:
        x (np.ndarray): values to remove outliers
        mask (np.ndarray): boolean mask where True is consider as an outlier
        axis (int, optional): outliers are removed through this dimension,
            returning a list instead of an array. Defaults to None.

    Returns:
        np.ndarray or list: the masked values
    """
    assert values.shape == mask.shape, "Both values and mask must have same shape"
    if axis is not None:
        result = []
        for idx in range(values.shape[axis]):
            result.append(np.take(values, idx, axis=axis)[np.take(mask, idx, axis=axis)])
        return result
    return values[mask]


def compute_saturation(x, /, axis=None, min_value=None, max_value=None):
    """Returns the percentage of saturating values.

    Args:
        x (np.ndarray or QTensor): the values to verify.
        axis (int, optional): saturation is computed through this dimension,
            returning a list instead of an array. Defaults to None.
        min_value (np.ndarray, optional): the minimum of values.
            If not provided, it is inferred from the values type. Defaults to None.
        max_value (np.ndarray, optional): the maximum of values.
            If not provided, it is inferred from the values type. Defaults to None.

    Returns:
        np.ndarray or list: the saturation percentage
    """
    try:
        # Overwrit max_value, min_value if x is a QTensor
        min_value = -2**x.value_bits
        max_value = 2**x.value_bits - 1
        x = x.values.numpy()
    except AttributeError:
        pass
    if axis is not None:
        saturation = []
        for idx in range(x.shape[axis]):
            metric = Saturation(min_value=min_value, max_value=max_value, dtype=x.dtype)
            # Compute saturation over axis
            metric.update_state(np.take(x, idx, axis=axis))
            saturation.append(metric)
        return saturation
    # Compute saturation over x
    metric = Saturation(min_value=min_value, max_value=max_value, dtype=x.dtype)
    metric.update_state(x)
    return metric


@make_fn_on_list
def compare(x, y, saturation, /):
    """Compare the set (x, y) element-wise

    Args:
        x (np.ndarray or list): the ground truth.
        y (np.ndarray or list): the array to compare with x.
        saturation (Saturation or list): saturation metric to be included in the result.

    Returns:
        dict or list: list of measures
    """
    smape = SMAPE()
    smape.update_state(x, y)
    return {"SMAPE": smape, "Saturation (%)": saturation}


@make_fn_on_list
@make_fn_on_dict
def eval_metrics(metric, /):
    """Call 'result' function in a set of metrics

    Args:
        metric (list or dict or keras.Metric): the metric to be evaluated

    Returns:
        list or dict or np.ndarray: the evaluated metrics
    """
    return metric.result().numpy()



#!/usr/bin/env python
# ******************************************************************************
# Copyright 2023 Brainchip Holdings Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ******************************************************************************
"""
Multiple Reshape/Flatten transformations
"""

__all__ = ["remove_reshape"]

from keras.layers import Reshape, Flatten
from keras.models import Sequential
from copy import deepcopy
from .transforms_utils import get_layers, get_layer_index, update_inbound


def _find_reshape_flatten_sequence(model):
    """ Retrieves a Reshape/Flatten layer that is followed by another

    Args:
        model (keras.Model): the model

    Returns:
        keras.layer: a Reshape or Flatten layer that is followed by another. Or None if not found.
    """
    expected_type = (Reshape, Flatten)
    for layer in model.layers:
        # Find candidate layer and get the type of the next layer if there is one
        if isinstance(layer, expected_type) and len(layer.outbound_nodes) == 1:
            outbound = layer.outbound_nodes[0].layer
            if isinstance(outbound, expected_type):
                return layer
    return None


def _get_layer_less_model(model, layer_del):
    """ Edits the model configuration to remove the layer_del and rebuilds a model.

    Args:
        model (keras.Model): the model
        layer_del (keras.layer): the layer to delete

    Returns:
        keras.Model: the updated model without the layer to delete
    """
    # get_config documentation mentions that a copy should be made when planning to modify the
    # config.
    config = deepcopy(model.get_config())
    layers = config['layers']

    # For sequential model, the changes stop here: the layer will simply be removed in the
    # following step. For other models, the layers inbounds/outbounds must be rebuilt.
    if not isinstance(model, Sequential):
        del_index = get_layer_index(layers, layer_del.name)
        # tfmot code: 'inbound_nodes' is a nested list where first element is the inbound layername,
        # e.g: [[['conv1', 0, 0, {} ]]]
        updated_inbound = layers[del_index]['inbound_nodes'][0][0][0]

        # Get the layers after del_index, ie. outbounds layers
        del_outbound_names = [outbound.layer.name for outbound in layer_del.outbound_nodes]
        outbound_ids = [
            get_layer_index(layers, del_outbound) for del_outbound in del_outbound_names]

        # "inbound > layer_del > outbounds" becomes "inbound > outbounds"
        for id in outbound_ids:
            update_inbound(layers[id], layer_del.name, updated_inbound)

    # Remove the layer
    layers_to_remove = get_layers(config, layer_del.name)
    assert len(layers_to_remove) == 1
    layers.remove(layers_to_remove[0])

    # Reconstruct model from the config, using the cloned layers
    updated_model = model.from_config(config)

    return updated_model


def remove_reshape(model):
    """ Multiple Reshape/Flatten transformations

    The patterns are:
    - Reshape + Reshape → Reshape
    - Flatten + Flatten → Flatten
    - Reshape + Flatten → Flatten
    - Flatten + Reshape → Reshape

    Args:
        model (keras.Model): a model

    Returns:
        keras.Model: the original model or the updated model if needed
    """
    # We consider successively the following sequences of layers:
    # Reshape > Reshape or Reshape > Flatten or Flatten > Reshape or Flatten > Flatten
    updated_model = model
    while True:
        # Find a layer that matches with the layers pattern
        layers_to_delete = _find_reshape_flatten_sequence(updated_model)
        if layers_to_delete is None:
            break  # No more layer to delete

        # We remove the selected layer
        updated_model = _get_layer_less_model(updated_model, layers_to_delete)

    # Restore model weights if there is a transformation
    if model != updated_model:
        updated_model.set_weights(model.get_weights())

    return updated_model



#!/usr/bin/env python
# ******************************************************************************
# Copyright 2022 Brainchip Holdings Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ******************************************************************************

__all__ = ["OutputQuantizer"]

import tensorflow as tf
import keras

from ...tensors import QTensor, FixedPoint, QFloat
from ..recorders import TensorRecorder, FixedPointRecorder
from .quantizers import Quantizer


@keras.saving.register_keras_serializable()
class OutputQuantizer(Quantizer):
    """A uniform FixedPoint quantizer that selects the optimal number of fractional bits for the
    range of its inputs and updates them accordingly.

    The typical use case is to decrease the bitwidth of the result of a quantized layer operation to
    avoid a saturation in downstream operations.

    If the input is a QFloat, it is converted to a FixedPoint before updating its bitwidth.

    Args:
        bitwidth (int, optional): the quantization bitwidth. Defaults to 8.
        signed (bool, optional): whether the quantizer expects signed values or unsigned.
            Defaults to True.
        axis (str, optional): the quantization range is a scalar ('per-tensor') or a vector
            corresponding to the last axis ('per-axis'). Defaults to 'per-tensor'.
        scale_bits: (int, optional): the bitwidth to use when quantizing output scales.
            Defaults to 8.
        buffer_bitwidth: (int, optional): buffer bitwidth value. Defaults to 32.
    """

    def __init__(self,
                 bitwidth=8,
                 signed=True,
                 axis="per-tensor",
                 scale_bits=8,
                 buffer_bitwidth=32,
                 **kwargs):
        super().__init__(bitwidth, signed, **kwargs)
        if not (isinstance(axis, str) and axis in ["per-tensor", "per-axis"]):
            raise ValueError(f"Only support reduction 'per-tensor' or 'per-axis'. Given {axis}.")
        self.axis = axis
        self.scale_bits = scale_bits
        self.buffer_bitwidth = buffer_bitwidth
        # Add object that will store the shift values.
        self.shift = TensorRecorder(self.name + "/shift")

    def build(self, input_shape):
        """Build the layer.

        Args:
            input_shape (list): the shape of input tensor.
        """
        super().build(input_shape)
        # Convert axis to a list of int
        if self.axis == "per-axis":
            ndims = len(input_shape)
            if ndims < 3:
                raise ValueError("OutputQuantizer cannot quantize per-axis tensors "
                                 " with 2 dimensions or less.")
            self._axis = list(range(len(input_shape) - 1))
        else:
            self._axis = None

        # Declares the constant/vector that will store the maximum values of the input.
        self.range_max = self.add_weight(
            name="range_max",
            shape=input_shape[-1] if self._axis is not None else (),
            dtype=tf.float32,
            initializer="ones",
            synchronization=tf.VariableSynchronization.ON_READ,
            trainable=False,
            aggregation=tf.VariableAggregation.MEAN,
            experimental_autocast=False,
        )

        # Declare a rescaling_rate variable that will be set at calibration and that will hold
        # cross-layer equalization ideal range_max / calibrated range_max ratio.
        self.rescaling_rate = self.add_weight(
            name="rescaling_rate",
            shape=input_shape[-1] if self._axis is not None else (),
            dtype=tf.float32,
            initializer="ones",
            synchronization=tf.VariableSynchronization.ON_READ,
            trainable=False,
            aggregation=tf.VariableAggregation.MEAN
        )

    def call(self, inputs):
        """Quantize the QTensor inputs to a lower bitwidth.

        The quantization happens with the following steps:

            1. Evaluate the nearest power(s) of two containing the quantization range(s)
            2. Quantize the inputs.

        Args:
            inputs (:obj:`QTensor`): the inputs tensor.

        Returns:
            :obj:`FixedPoint`: the quantized tensor.
        """
        if not isinstance(inputs, QTensor):
            raise TypeError("The OutputQuantizer accepts only QTensor inputs."
                            f"Received {type(inputs)} inputs.")

        if isinstance(inputs, QFloat):
            if self.scale_bits is None:
                raise ValueError(f"{self.name} receives QFloat inputs: the scale_bits parameter"
                                 " needs to be specified.")
            inputs_value_bits = inputs.fp.value_bits
            # Apply cross layer equalization rescaling
            rescaled_inputs = QFloat(inputs.fp, inputs.scales * self.rescaling_rate)
            rescaled_inputs = rescaled_inputs.promote(self.buffer_bitwidth)
            inputs, qscales = rescaled_inputs.to_fixed_point(self.scale_bits)
            if getattr(self, 'qscales', None) is None:
                # From Keras documentation, any variable creation taking place in call
                # should be wrapped with tf.init_scope
                with tf.init_scope():
                    self.qscales = FixedPointRecorder(self.name + "/qscales")
            self.qscales(qscales)
        else:
            inputs_value_bits = inputs.value_bits

        if inputs_value_bits <= self.value_bits:
            msg = f"Quantizing a {inputs_value_bits}-bit QTensor to "\
                f"{self.value_bits}-bit is pointless."
            if inputs_value_bits < self.value_bits:
                msg += " Use a promotion instead."
            raise ValueError(msg)

        # Rescale to center around range_max and compress to a lower bitwidth
        frac_bits = tf.stop_gradient(FixedPoint.max_frac_bits(self.value_bits, self.range_max))
        inputs, shift_value = inputs.rescale(frac_bits, self.value_bits)
        # update shift values
        self.shift(shift_value)
        return inputs

    def get_config(self):
        """Get the config of the layer.

        Returns:
            dict: the config of the layer.
        """
        config = super().get_config()
        config.update({"scale_bits": self.scale_bits})
        config.update({"axis": self.axis})
        config.update({"buffer_bitwidth": self.buffer_bitwidth})
        return config



#!/usr/bin/env python
# ******************************************************************************
# Copyright 2023 Brainchip Holdings Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ******************************************************************************

__all__ = ["QuantizedBatchNormalization"]

import tensorflow as tf
import keras

from .layers_base import (register_quantize_target, rescale_outputs, register_aligned_inputs,
                          tensor_inputs, apply_buffer_bitwidth, QuantizedLayer)
from .quantizers import WeightQuantizer, AlignedWeightQuantizer, OutputQuantizer
from ..tensors import QTensor


def _validate_axis(axis, input_shape):
    """Validate an axis value and returns its standardized form.

    Args:
        axis (int, list, tuple): Value to validate. Can be an integer or a list/tuple of integers.
            Integers may be negative.
        input_shape (tuple): Reference input shape that the axis/axes refer to.

    Returns:
       list: normalized form of `axis`, i.e. a list with all-positive values.
    """
    input_shape = tf.TensorShape(input_shape)
    rank = input_shape.rank
    if not rank:
        raise ValueError(f"Input has undefined rank. Received: input_shape={input_shape}")

    # Convert axis to list and resolve negatives
    if isinstance(axis, int):
        axis = [axis]
    else:
        axis = list(axis)
    for idx, x in enumerate(axis):
        if x < 0:
            axis[idx] = rank + x

    # Validate axes
    for x in axis:
        if x < 0 or x >= rank:
            raise ValueError("Invalid value for `axis` argument. Expected 0 <= axis < inputs.rank "
                             f"(with inputs.rank={rank}). Received: axis={tuple(axis)}")
    if len(axis) != len(set(axis)):
        raise ValueError(f"Duplicate axis: {tuple(axis)}")
    return axis


@register_quantize_target(keras.layers.BatchNormalization)
@register_aligned_inputs
@keras.saving.register_keras_serializable()
class QuantizedBatchNormalization(QuantizedLayer):
    r"""Layer that normalizes its inputs, on the last axis.

    The normalization is applied like this:

    .. math::

        y = \\frac{(x - \\mu) \\cdot \\gamma}{\\sigma} + \\beta \\
            = \\frac{x \\cdot \\gamma}{\\sigma} - \\
              \\frac{\\mu\\cdot \\gamma}{\\gamma} + \\beta

    if we consider:

    .. math:: a = \\frac{\\gamma}{\\sigma}

    and

    .. math:: b = -\\frac{\\mu\\cdot \\gamma}{\\sigma} + \\beta


    The normalization can be re-written as:

    .. math:: y = a \\cdot x + b

    Note that this layer will hold variables with names gamma, beta, moving_mean (:math:`\\mu`),
    and moving_variance (:math:`\\sigma = \\sqrt{moving\_variance + \\epsilon}`), so they can be
    converted from a BatchNormalization layer. However, it's a and b that are going to be quantized.

    Args:
        quant_config (dict, optional): the serialized quantization configuration. Defaults to None.
        axis (int, optional): The axis that was normalized on the
            BatchNormalization layer. The only supported value is the
            last dimension.
        epsilon (float, optional): Small value to avoid dividing by zero.
            Defaults to 1e-3.
    """

    ignored_args = ["momentum",
                    "center",
                    "scale",
                    "beta_initializer",
                    "gamma_initializer",
                    "moving_mean_initializer",
                    "moving_variance_initializer",
                    "beta_regularizer",
                    "gamma_regularizer",
                    "beta_constraint",
                    "gamma_constraint",
                    "renorm",
                    "renorm_clipping",
                    "renorm_momentum",
                    "fused",
                    "trainable",
                    "virtual_batch_size",
                    "adjustment"
                    ]

    def __init__(self,
                 *args,
                 quant_config=None,
                 axis=-1,
                 epsilon=1e-3,
                 **kwargs):
        super().__init__(*args, quant_config=quant_config, **kwargs)
        out_quant_cfg = self.quant_config.get("output_quantizer", False)
        if out_quant_cfg:
            self.out_quantizer = OutputQuantizer(
                name="output_quantizer", **out_quant_cfg)
        else:
            self.out_quantizer = None
        if "a_quantizer" not in self.quant_config:
            self.quant_config["a_quantizer"] = {"bitwidth": 8}
        a_quantizer_cfg = self.quant_config["a_quantizer"]
        self.a_quantizer = WeightQuantizer(name="a_quantizer", **a_quantizer_cfg)
        b_quantizer_cfg = self.quant_config.get("b_quantizer", {})
        self.b_quantizer = AlignedWeightQuantizer(name="b_quantizer", **b_quantizer_cfg)
        self.buffer_bitwidth = apply_buffer_bitwidth(self.quant_config, signed=True)

        # Define a small float number to avoid dividing by zero.
        self.epsilon = epsilon
        # Axis on which operation is applied
        self.axis = axis

    def build(self, input_shape):
        input_shape = tf.TensorShape(input_shape)
        rank = input_shape.rank

        if rank not in (3, 4):
            raise ValueError(
                "QuantizedBatchNormalization only supports 3D or 4D tensors. "
                f"Received tensor with shape: {tuple(input_shape)}.")

        # Normalize axis
        self.axis = _validate_axis(self.axis, input_shape)

        # Check selected axis is valid
        if len(self.axis) != 1 and (self.axis[0] != rank - 1):
            raise ValueError("QuantizedBatchNormalization only supports axis "
                             "argument set to the last dimension.")

        # Shape for variables is always as if it was applied on the
        # last dimension.
        param_shape = input_shape[-1]

        # Add BN compatible weights
        # Gamma
        self.gamma = self.add_weight(
            name="gamma",
            shape=param_shape,
            dtype=tf.float32,
            initializer="ones",
            regularizer=None,
            constraint=None,
            trainable=True,
            experimental_autocast=False,
        )
        # Beta
        self.beta = self.add_weight(
            name="beta",
            shape=param_shape,
            dtype=tf.float32,
            initializer="zeros",
            regularizer=None,
            constraint=None,
            trainable=True,
            experimental_autocast=False,
        )
        # Mu = moving mean
        self.moving_mean = self.add_weight(
            name="moving_mean",
            shape=param_shape,
            dtype=tf.float32,
            initializer="zeros",
            regularizer=None,
            constraint=None,
            trainable=False,
            experimental_autocast=False,
        )
        # Sigma² = moving variance
        self.moving_variance = self.add_weight(
            name="moving_variance",
            shape=param_shape,
            dtype=tf.float32,
            initializer="ones",
            regularizer=None,
            constraint=None,
            trainable=False,
            experimental_autocast=False,
        )

    @property
    def sigma_rec(self):
        # Sigma reciprocal = 1 / sigma = 1 / sqrt(moving_variance + epsilon)
        sigma_rec = tf.math.rsqrt(self.moving_variance + self.epsilon)
        return sigma_rec

    @property
    def a(self):
        a_var = self.gamma * self.sigma_rec
        q_a = self.a_quantizer(a_var)
        return q_a

    def b(self, inputs):
        sigma_rec = self.sigma_rec
        b_var = -self.moving_mean * self.gamma * sigma_rec + self.beta
        q_b = self.b_quantizer(b_var, inputs)
        return q_b

    @tensor_inputs([QTensor])
    @rescale_outputs
    def call(self, inputs):
        # Calculation is equivalent to
        # y = (x - mu) * gamma / sigma + beta
        #   = x * gamma / sigma - mu * gamma / sigma + beta
        #
        # So if we consider
        # a = gamma / sigma
        # b = -mu * gamma / sigma + beta
        # Then the evaluation is just y = a * x + b.

        # outputs = a * x
        outputs = tf.multiply(inputs, self.a)

        # quantize and retrieve b, aligned on the outputs to allow sum
        b = self.b(outputs)
        # y = outputs + b
        return tf.add(outputs, b)

    def get_config(self):
        config = super().get_config()
        config.update({"epsilon": self.epsilon, "axis": self.axis})
        return config



#!/usr/bin/env python
# ******************************************************************************
# Copyright 2022 Brainchip Holdings Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ******************************************************************************
"""
Transformation to fold BatchNormalization layers in a model.
"""

__all__ = ['fold_batchnorms']

import keras
import numpy as np
import tensorflow as tf

from copy import deepcopy
from keras.layers import (BatchNormalization, Conv2D, SeparableConv2D, Dense, DepthwiseConv2D,
                          Conv2DTranspose)

from .transforms_utils import get_layers, get_layer_index, update_inbound, find_layers_pairs
from ..utils import apply_weights_to_model
from ...layers import (PaddedConv2D, QuantizedConv2D, QuantizedSeparableConv2D, QuantizedDense,
                       QuantizedDepthwiseConv2D, QuantizedConv2DTranspose, DepthwiseConv2DTranspose,
                       QuantizedDepthwiseConv2DTranspose, BufferTempConv, DepthwiseBufferTempConv)


def fold_batchnorms(model):
    """Returns a new model where BatchNormalization layers are folded into previous layers.

    From a Keras model where BN layers follow processing layers, this function removes the BN layers
    and updates the preceding layers weights and bias accordingly. The new model is strictly
    equivalent to the previous one.

    Args:
        model (keras.Model): a model

    Returns:
        keras.Model: the original model or a model with BN folded
    """
    # Find foldable BN and their inbound
    map_prev_layer_to_bn = _find_batchnorms_to_fold(model)

    # When there are no valid candidates, return the original model
    if len(map_prev_layer_to_bn) == 0:
        return model

    # Rebuild a model without BN by editing the configuration
    model_folded = _get_BN_less_model(model, map_prev_layer_to_bn)

    # Set folded weights in the new model
    for layer, bn in map_prev_layer_to_bn.items():
        new_weights = _compute_BN_folded_weights(layer, bn)
        model_folded.get_layer(layer.name).set_weights(new_weights)

    return model_folded


def _find_batchnorms_to_fold(model):
    """ Retrieves BatchNormalization layers that can be folded.

    This is limited to BN layers that follow supported layer types.

    Args:
        model (keras.Model): a model

    Returns:
        dict: map between a layer and its BN layer to be folded

    Raises:
        RuntimeError: if the candidate BN has null gammas or BN is not applied on the last axis.
    """

    supported_layers = (Conv2D, PaddedConv2D, Conv2DTranspose, SeparableConv2D, DepthwiseConv2D,
                        Dense, QuantizedConv2D, QuantizedConv2DTranspose, QuantizedSeparableConv2D,
                        QuantizedDepthwiseConv2D, QuantizedDense, DepthwiseConv2DTranspose,
                        QuantizedDepthwiseConv2DTranspose, BufferTempConv, DepthwiseBufferTempConv)

    # Map between a layer and its following BN layer
    map_prev_layer_to_bn = find_layers_pairs(model, supported_layers, BatchNormalization)

    # Check valid "layer -> BN" candidates
    for bn in map_prev_layer_to_bn.values():
        # Null gammas are not supported: once folded, new kernel would be zero
        gamma = bn.get_weights()[0]
        if np.any(gamma == 0):
            raise RuntimeError(f"There are {np.sum(gamma == 0)} null gammas in the "
                               f"BatchNormalization layer '{bn.name}'. Null gammas are "
                               "not supported.")

        # Check BN axis parameter
        if (len(bn.axis) != 1 or bn.axis[0] != len(bn.input_shape) - 1):
            raise RuntimeError(f"The BatchNormalization must be applied on the last "
                               f"axis. '{bn.name}' layer has axis={bn.axis}.")
    return map_prev_layer_to_bn


def _get_BN_less_model(model, map_prev_layer_to_bn):
    """ Edits the model configuration to remove BN layer and rebuilds a model.

    Args:
        model (keras.Model): a model
        map_prev_layer_to_bn (dict): map between a layer and its BN layer to be folded

    Returns:
        keras.Model: an updated model without BN layers
    """
    # get_config documentation mentions that a copy should be made when planning to modify the
    # config
    config = deepcopy(model.get_config())
    layers = config['layers']
    new_output_name = None

    for prev_layer, layer_bn in map_prev_layer_to_bn.items():
        # Set use_bias=True in layers where BN are folded to accept new computed weights
        prev_index = get_layer_index(layers, prev_layer.name)
        layers[prev_index]['config']['use_bias'] = True

        # For sequential model, the changes stop here: the BN layers will simply be removed in the
        # following step. For other models, the layers inbounds/outbounds must be rebuilt.
        if isinstance(model, keras.Sequential):
            continue

        # Retrieve the BN input layer. Assumes that BN has only 1 inbound.
        bn_index = get_layer_index(layers, layer_bn.name)
        # tfmot code: 'inbound_nodes' is a nested list where first element is the inbound layername,
        # e.g: [[['conv1', 0, 0, {} ]]]
        updated_inbound = layers[bn_index]['inbound_nodes'][0][0][0]

        # Get the layers after BN, ie. outbounds layers
        bn_outbound_names = [outbound.layer.name for outbound in layer_bn.outbound_nodes]
        outbound_ids = [get_layer_index(layers, bn_outbound) for bn_outbound in bn_outbound_names]

        # If BN is the last layer (no outbound), store its name and inbound name so that the model
        # output can be updated later
        if len(outbound_ids) == 0:
            new_output_name = updated_inbound
            last_bn = layer_bn.name

        # Update BN outbounds layers: their current inbound is the BN layer that will be folded so
        # it must be replaced with the BN previous layer (folding target or the BN inbound). This
        # results in by-passing the BN layer: inbound > BN > outbounds becomes inbound > outbounds.
        for id in outbound_ids:
            update_inbound(layers[id], layer_bn.name, updated_inbound)

    # Remove BN layers
    layers_to_remove = get_layers(config, [bn.name for bn in map_prev_layer_to_bn.values()])
    for layer_to_remove in layers_to_remove:
        layers.remove(layer_to_remove)

    # Update the model outputs if needed
    if new_output_name:
        for index, out_layer in enumerate(config['output_layers']):
            if out_layer[0] == last_bn:
                config['output_layers'][index][0] = new_output_name

    # Reconstruct model from the config, using the cloned layers.
    updated_model = model.from_config(config)

    # Restore model weights
    variables_dict = {var.name: var for var in model.variables}
    apply_weights_to_model(updated_model, variables_dict, False)
    return updated_model


def _compute_BN_folded_weights(target_layer, bn_layer):
    """ Computes the new weights of a layer after folding BatchNormalization into it.

    Args:
        target_layer (keras.layers.Layer): the layer where BN will be folded
        bn_layer (keras.layers.Layer): the BatchNormalization layer to fold into the preceding layer

    Returns:
        list: the new weights to set in the folded destination layer.
    """
    # Get kernel and bias weights of the target layer
    kernel_position = 0
    bias_position = 1
    if isinstance(target_layer, SeparableConv2D):
        kernel_position = 1
        bias_position = 2

    weights = target_layer.get_weights()
    kernel = weights[kernel_position]
    bias = weights[bias_position] if target_layer.use_bias else 0

    # Get BN weights
    gamma, beta, mean, var = bn_layer.get_weights()
    scale_BN = gamma / np.sqrt(var + bn_layer.epsilon)

    # Compute new folded kernel and bias
    if isinstance(target_layer, DepthwiseConv2D):
        # Last dimension of DepthwiseConv2D kernel is 1 so scale_BN should be reshaped to (shape, 1)
        # (first two dimensions will be broadcasted)
        new_kernel = kernel * scale_BN.reshape(*scale_BN.shape, 1)
    elif isinstance(target_layer, Conv2DTranspose):
        # Handle Conv2DTranspose kernels that are (h, w, filters, channels) instead of (h, w,
        # channels, filters)
        new_kernel = tf.transpose(kernel, (0, 1, 3, 2)) * scale_BN
        new_kernel = tf.transpose(new_kernel, (0, 1, 3, 2))
    else:
        new_kernel = kernel * scale_BN
    new_bias = beta + (bias - mean) * scale_BN

    # Return all weights with modified ones
    new_weights = weights
    new_weights[kernel_position] = new_kernel
    if target_layer.use_bias:
        new_weights[bias_position] = new_bias
    else:
        new_weights.append(new_bias)

    return new_weights



#!/usr/bin/env python
# ******************************************************************************
# Copyright 2023 Brainchip Holdings Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ******************************************************************************
""" Transformation to invert BatchNormalization and Pooling layers in a model."""

__all__ = ['invert_batchnorm_pooling']

import numpy as np

from keras.layers import BatchNormalization, MaxPool2D, GlobalAvgPool2D

from .transforms_utils import find_layers_pairs, invert_layer_pairs


def invert_batchnorm_pooling(model):
    """ Inverts pooling and BatchNormalization layers in a model to have BN layer before pooling.

    Returns a new model where pooling and batch normalization layers are inverted. From a Keras
    model where pooling layers precede batch normalization layers, this function places the BN
    layers before pooling layers. This is the first step before folding BN layers into processing
    layers.

    Note:
        Inversion of layers is equivalent only if the gammas of BN layers are positive. The
        function raises an error if not.

    Args:
        model (keras.Model): a model

    Returns:
        keras.Model: the updated model

    Raises:
        RuntimeError: if a candidate BatchNormalization layer has gamma values that are not strictly
            positive.
    """
    # Find pooling followed by BN layer pairs that are candidates for inversion
    map_pool_bn = find_layers_pairs(model, (MaxPool2D, GlobalAvgPool2D), BatchNormalization)

    # When there are no valid candidates, return the original model
    if not map_pool_bn:
        return model

    # Before inverting, check gamma values and update BN axis with GAP
    for pool, bn in map_pool_bn.items():
        gammas = bn.get_weights()[0]
        if isinstance(pool, MaxPool2D) and np.any(gammas <= 0):
            # It is impossible to invert MaxPool->BN with gammas <= 0
            raise RuntimeError(f"There are {np.sum(gammas <= 0)} negative gammas in the "
                               f"BatchNormalization layer {bn.name}. Negative gammas are "
                               "not supported.")
        if isinstance(pool, GlobalAvgPool2D):
            # Update BN axis when preceeded by GAP
            bn.axis = [-1]

    # Rebuild a model with pooling and BN inverted by editing the configuration
    return invert_layer_pairs(model, map_pool_bn)



#!/usr/bin/env python
# ******************************************************************************
# Copyright 2023 Brainchip Holdings Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ******************************************************************************
__all__ = ["InputQuantizer", "Dequantizer", "get_input_quantizer"]

import numpy as np

from onnx.helper import make_node
from onnx import AttributeProto as AP
from onnxruntime.quantization.calibrate import TensorData

from .base_layer import OnnxLayer
from .layer_compatibility import check_node_link_to_input, check_node_has_one_initializer
from ..graph_tools import (TENSOR_SHAPE, value_info_to_tensor_shape, array_to_tp, get_field,
                           get_variable, get_node, has_field, to_field)

from ..quantization.input_scale import input_zp_scale, needs_zp


def get_input_quantizer(nodes, graph):
    check_node_link_to_input(nodes[0], graph)

    perm = None
    transpose_node = get_node(nodes, "Transpose")
    if (transpose_node := get_node(nodes, "Transpose")) is not None:
        perm = get_field(transpose_node, "perm")

    input_ts = graph.input[0]
    input_unsigned = needs_zp(graph, nodes[-1].output[0])
    input_quantizer = InputQuantizer(name="quantize",
                                     input_tp=input_ts,
                                     input_signed=not input_unsigned,
                                     perm=perm)

    # Sets rescaling weights
    if (mul_node := get_node(nodes, "Mul")) is not None:
        check_node_has_one_initializer(mul_node, graph)
        input_quantizer.set_weight("input_scale", get_variable(mul_node.input[1], graph))
    if (add_node := get_node(nodes, "Add")) is not None:
        check_node_has_one_initializer(add_node, graph)
        input_quantizer.set_weight("offset", get_variable(add_node.input[1], graph))
    return input_quantizer


def _compute_quantization_parameters(r_scale, r_offset, out_tensor_range, signed):
    assert isinstance(out_tensor_range, TensorData)
    # Computing the ranges before the Rescale node (e.g. the model inputs)

    in_tensor_range = [(x - r_offset) / r_scale for x in out_tensor_range.range_value]
    in_tensor_range = TensorData(lowest=np.minimum(*in_tensor_range),
                                 highest=np.maximum(*in_tensor_range))

    # Compute scale and zero point to quantize the inputs
    input_scale, input_zp = input_zp_scale(in_tensor_range, allow_zp=not signed)

    # Compute scale and output to dequantize the outputs
    output_scale, output_zp = input_zp_scale(out_tensor_range, allow_zp=not signed)

    # Check constraints
    err_msg = "Impossible to quantize inputs when folding rescale: "
    if not signed:
        # Compare if rescale is valid
        np.testing.assert_allclose(input_scale / output_scale,
                                   r_scale,
                                   atol=1e-3,
                                   err_msg=err_msg + "input/output scales ratio is not valid.")
    else:
        np.testing.assert_equal(r_offset, 0, err_msg + "offset must be zero when input is signed.")
    return (input_scale, input_zp), (output_scale, output_zp)


class InputQuantizer(OnnxLayer):
    """Intermediate representation of QuantizeLinear(), use to quantize the input.

    Args:
        input_tp (TensorProto): the input of the ONNX model.
        perm (list, optional): list representing the permutations of the rescale node.
            Defaults to None.
        input_signed (bool, optional): whether the input is signed. Defaults to False.
        name (str, optional): the node name. Defaults to ''.
    """

    def __init__(self, input_tp, perm=None, input_signed=False, name=''):
        super().__init__("InputQuantizer", name=name, perm=perm)
        self.input_signed = input_signed
        self._input = [input_tp]

        # Declare weights
        self._add_weight("input_scale")
        self._add_weight("offset")

    def __build__(self, downscale=True):
        assert downscale, f"{self.name} ({self.base_name}) does not support 32bit output"
        input_ts = value_info_to_tensor_shape(self.input)
        assert input_ts.dtype == np.float32

        # Add/initialize weights
        zp_dtype = "int8" if self.input_signed else "uint8"
        self._add_weight("zero_point", value=np.zeros(input_ts.shape[1]), dtype=zp_dtype)
        if self.weights["input_scale"].size == 0:
            self.set_weight("input_scale", np.ones((), dtype="float32"))
        if self.weights["offset"].size == 0:
            self.set_weight("offset", np.zeros((), dtype="float32"))

        # Update perm attribute
        input_ndim = len(input_ts.shape)
        if has_field(self, "perm"):
            perm = get_field(self, "perm")
        else:
            perm = list(range(input_ndim))
            self.attribute.append(to_field("perm", perm))

        # Compute output shape
        output_shape = tuple(input_ts.shape[i] for i in perm)
        output_ts = TENSOR_SHAPE(output_shape, np.dtype(zp_dtype))

        # Assert wrong weights format
        first_channel_expected_shape = (1,) + output_shape[1:2] + (1,) * (input_ndim - 2)
        if (self.weights["input_scale"].size != 1 and
                self.weights["input_scale"].shape != first_channel_expected_shape):
            raise ValueError(f"Unsupported 'input_scale' in {self.name} ({self.base_name}): "
                             "it must be brodcastable in the channels dimension.")
        if (self.weights["offset"].size != 1 and
                self.weights["offset"].shape != first_channel_expected_shape):
            raise ValueError(f"Unsupported 'offset' in {self.name} ({self.base_name}): "
                             "it must be brodcastable in the channels dimension.")
        return output_ts

    def __quantize__(self, out_tensor_range, force_fp=False):
        if force_fp:
            raise NotImplementedError("Force input scale to be a FP is not implemented yet.")

        # Calibration was done by axis=1. Therefore we can squeeze dimension in mean/offset
        rescale_scale = np.squeeze(self.weights["input_scale"])
        rescale_offset = np.squeeze(self.weights["offset"])

        # Computing quantization parameters
        input_scale_zp, output_scale_zp = _compute_quantization_parameters(rescale_scale,
                                                                           rescale_offset,
                                                                           out_tensor_range,
                                                                           self.input_signed)

        # Scale to set in weights is the reciprocal of ONNX calibrated one.
        input_scale = np.array(1 / input_scale_zp[0], dtype=np.float32)

        # Compute weights to serialize
        weights = {f"{self.name}_scale": input_scale, f"{self.name}_zp": input_scale_zp[1]}

        # Save zero point (used by next layer)
        self.set_weight("zero_point", output_scale_zp[1])
        return weights, output_scale_zp[0]

    @staticmethod
    def build_subgraph(op_type):
        nodes = [make_node('Transpose', inputs=["X"], outputs=["Yi"])]
        nodes[-1].attribute.append(AP(name="perm", ref_attr_name="perm", type=AP.INTS))
        nodes.append(make_node('QuantizeLinear', inputs=["Yi", "scale", "zp"], outputs=["Y"]))
        return nodes


class Dequantizer(OnnxLayer):
    """Intermediate representation of DequantizeLinear(), use to dequantize the input.

    Args:
        name (str, optional): the node name. Defaults to ''.
    """

    def __init__(self, name=''):
        super().__init__("Dequantizer", name=name)

    def __build__(self, input_ts, downscale=True):
        assert input_ts.dtype in (np.int8, np.int32)

        # Compute output shape
        output_ts = TENSOR_SHAPE(input_ts.shape, np.dtype("float32"))
        return output_ts

    def quantize(self, qinput):
        # To keep homogenity with the other layers, this function is called 'quantize'
        # even though it does the opposite (dequantize): apply scale in the input integers.
        if self._output is None or self._input is None:
            # Build the layer if required
            self.build(qinput.output)

        # Scale to set in weights is the reciprocal of ONNX calibrated one.
        i_scale = qinput.weights["scale"]
        scale = np.array(1 / i_scale, dtype=np.float32)

        # Return ONNX node and weights
        weights = {f"{self.name}_scale": scale}
        inputs = [ts.name for ts in self._input] + list(weights)
        onnx_node = self.make_node(inputs, [self.output.name])
        onnx_weights = array_to_tp(**weights)
        return onnx_node, onnx_weights

    @staticmethod
    def build_subgraph(op_type):
        return [make_node('DequantizeLinear', inputs=["X", 'scale'], outputs=["Y"])]



#!/usr/bin/env python
# ******************************************************************************
# Copyright 2023 Brainchip Holdings Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ******************************************************************************
__all__ = ["cast_tensors_to"]

from onnx import TensorProto as TP
from onnx.helper import make_node


def cast_tensors_to(in_names, dtype=TP.FLOAT):
    """Cast tensors to a given type.

    Args:
        in_names (list): list of input tensor names.
        dtype (DataType, optional): the type to cast. Defaults to TensorProto.FLOAT

    Returns:
        NodeProto list, list: list of cast nodes and output names.
    """
    # Cast tensors to dtype
    nodes = []
    out_names = []
    for in_name in in_names:
        out_name = ""
        if in_name:
            # Cast every tensor with name
            out_name = f"{in_name}/cast"
            nodes.append(make_node("Cast", [in_name], [out_name], to=dtype))
        out_names.append(out_name)
    return nodes, out_names



#!/usr/bin/env python
# ******************************************************************************
# Copyright 2023 Brainchip Holdings Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ******************************************************************************
__all__ = ["has_field", "get_field", "to_field", "replace_field"]

import numpy as np

import onnx
import onnx.numpy_helper
from onnx.helper import make_attribute
from onnxruntime.quantization.quant_utils import find_by_name


def has_field(node, name):
    """Helper to check if a node has a field given its name.

    Args:
        node (NodeProto): the node to read the field.
        name (str): the name of the field.

    Returns:
        bool: if the node contains the attribute.
    """
    attr = find_by_name(name, node.attribute)
    return attr is not None


def get_field(node, name: str, default=None):
    """Helper to get the value of a field of a node.

    Args:
        node (NodeProto): the node to read the field.
        name (str): the name of the field.
        default (Any, optional): if the field is not found, return this value.
            If not provided, raise an exception if not field. Defaults to None.

    Returns:
        the value of the field as np.array.
    """
    attr = find_by_name(name, node.attribute)
    if attr is None:
        assert default is not None, f"Node {node.name} does not have attribute {name}."
        # onnx.helper.get_attribute_value converts any tuple/list into a list
        if isinstance(default, (tuple, list)):
            default = list(default)
        return default
    value = onnx.helper.get_attribute_value(attr)
    if isinstance(value, onnx.TensorProto):
        # Convert value into an array when is a TensorProto
        value = onnx.numpy_helper.to_array(value)
    elif isinstance(value, bytes):
        value = value.decode()
    return value


def to_field(name, value):
    """Helper to convert a value into an AttributeProto.

    Args:
        name (str): the attribute name.
        value (Any): the attribute value.

    Returns:
        AttributeProto: the attribute
    """
    if not isinstance(value, onnx.AttributeProto):
        # Convert any numpy array into a ProtoTensor
        if isinstance(value, np.ndarray):
            value = onnx.numpy_helper.from_array(value)
        value = make_attribute(name, value)
    else:
        # Try to read the value to know if it is a valid attribute
        onnx.helper.get_attribute_value(value)
        # And verify name is correct
        assert value.name == name
    return value


def replace_field(node, attr_name, new_value):
    """Helper to replace the value of one attribute in a node by another.

    Args:
        node (NodeProto): the node.
        attr_name (str): the attribute name of the value to be replaced.
        new_value (Any): the new value.
    """
    # Convert new value into an attribute
    new_value_attr = to_field(attr_name, new_value)

    # We replace field only if types are the same
    old_attribute = find_by_name(attr_name, node.attribute)
    assert old_attribute is not None, f"Node {node.name} does not have attribute {attr_name}."
    if old_attribute.type != new_value_attr.type:
        raise ValueError(f"Wrong attribute type. Expected {old_attribute.type}")

    # Replace information
    old_attribute.CopyFrom(new_value_attr)



#!/usr/bin/env python
# ******************************************************************************
# Copyright 2024 Brainchip Holdings Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ******************************************************************************

__all__ = ["convert_squeeze_reshape_to_flatten"]

import numpy as np

from ...graph_tools import get_tensor_shape, check_node_attributes
from ..model import ONNXModel


def convert_squeeze_reshape_to_flatten(model):
    """Converts Reshape or Squeeze nodes into a Flatten node when possible.

    Args:
        model (ONNXModel): The ONNX model to be processed.
    """
    assert isinstance(model, ONNXModel)

    expected_op_type = ("Reshape", "Squeeze")
    supported_attributes = {'allowzero': [0]}

    for node in model.nodes():
        if node.op_type in expected_op_type:
            # Reshape and Squeeze should have only one non initializer
            if len(model.get_node_inputs(node)) > 1:
                continue

            input_value_info = model.find_value_info_by_name(node.input[0])
            output_value_info = model.find_value_info_by_name(node.output[0])

            input_shape = get_tensor_shape(input_value_info)
            output_shape = get_tensor_shape(output_value_info)
            input_prod = np.prod(input_shape[1:])

            # Test if input shape flattened is equal to 2D output shape
            if (input_shape[0], input_prod) == output_shape:
                # Squeeze is not equivalent to Flatten if no axes are
                # specified, because in that case, if the batch size is 1,
                # Squeeze will remove it, unlike Flatten.
                if node.op_type == "Squeeze":
                    try:
                        axes = model.get_variable(node.input[1])
                        if 0 in axes:
                            # In case batch size is in the axis to be squeezed,
                            # we continue
                            continue
                    except IndexError:
                        # If no axes are provided, we continue
                        continue

                # Check node attributes
                try:
                    check_node_attributes(node, supported_attributes)
                except ValueError:
                    # We don't change the model when the node attributes
                    # do not match the required constraints
                    continue

                # Clean all attributes
                node.ClearField("attribute")

                # Change op to Flatten
                node.op_type = "Flatten"

                # Remove if there is a second input (e.g. shape/axes)
                if len(node.input) == 2:
                    node.input.pop(1)

    # Clean shape/axes from initializers
    model.clean_initializers()
    model.check_model()



#!/usr/bin/env python
# ******************************************************************************
# Copyright 2023 Brainchip Holdings Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ******************************************************************************
__all__ = ["get_op_version"]


import onnx


def get_op_version(op_type, model):
    """Helper to get the opset version of an operation type in the model.

    Args:
        op_type (str): the operation type
        model (ModelProto): the model containing the operation.

    Returns:
        int: the opset version
    """
    for opset_import in model.opset_import:
        if onnx.defs.has(op_type, opset_import.domain):
            return opset_import.version
    raise RuntimeError(f"Model does not contain a version for '{op_type}'.")



#!/usr/bin/env python
# ******************************************************************************
# Copyright 2023 Brainchip Holdings Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ******************************************************************************
__all__ = ['fold_pad_into_conv']

from ...graph_tools import get_field, replace_field, has_field, to_field
from ..model import ONNXModel


def fold_pad_into_conv(model):
    assert isinstance(model, ONNXModel)
    pre_msg_erro = "Impossible to fold {} into {}: "

    for tnode in model.nodes()[1:]:
        if tnode.op_type != 'Conv':
            continue
        pnode = model.get_parent(tnode, idx=0)
        if pnode is not None and pnode.op_type == 'Pad':
            # Pad node should have only one non initializer
            if len(model.get_node_inputs(pnode)) > 1:
                continue

            # Raise exception if pad has multiple outbounds
            node_outbounds = model.get_children(pnode)
            if len(node_outbounds) != 1:
                raise RuntimeError(pre_msg_erro.format(pnode.name, tnode.name) +
                                   f"Single outbound is required for {pnode.name} " +
                                   f"(found {[x.name for x in node_outbounds]}.")

            # Check valid model
            mode = get_field(pnode, "mode", "constant")
            if mode != "constant":
                raise RuntimeError(pre_msg_erro.format(pnode.name, tnode.name) +
                                   f"Unsupported '{mode}' mode.")

            # Check valid constant value
            if len(pnode.input) > 2 and pnode.input[2]:
                constant_value = model.get_variable(pnode.input[2])
                if constant_value != 0:
                    raise RuntimeError(pre_msg_erro.format(pnode.name, tnode.name) +
                                       "Constant value is not zero.")

            # Input rank in a 'Conv' is the same as kernel one
            input_ndim = model.get_variable(tnode.input[1]).ndim

            # Retrieve pad_pads and axes (if any)
            pad_pads = model.get_variable(pnode.input[1]).tolist()
            if len(pnode.input) > 3 and pnode.input[3]:
                axes = model.get_variable(pnode.input[3])
                axes = [x if x >= 0 else input_ndim + x for x in axes]
            else:
                axes = list(range(input_ndim))

            # Fulfill pad_pads in every dimension (filling with zero the other ones)
            for axis in range(input_ndim):
                if axis not in axes:
                    pad_len = len(pad_pads) // 2
                    pad_pads.insert(pad_len + axis, 0)
                    pad_pads.insert(axis, 0)

            # Check pad only in spatial dimensions
            if any(pad_pads[:2] + pad_pads[input_ndim:input_ndim + 2]):
                raise RuntimeError(pre_msg_erro.format(pnode.name, tnode.name) +
                                   "Pad has non-zero values on batch or channel dimension.")
            # Get only spatial pads
            new_pads = pad_pads[2:input_ndim] + pad_pads[input_ndim + 2:]

            # Replace conv pads = new + old
            if has_field(tnode, "pads"):
                target_pads = get_field(tnode, "pads", [0] * 2 * (input_ndim - 2))
                replace_field(tnode, "pads", [x + y for x, y in zip(target_pads, new_pads)])
            else:
                new_pads_attr = to_field("pads", new_pads)
                tnode.attribute.append(new_pads_attr)

            # Prune node
            model.remove_node(pnode, update_graph=True)

    # Clean graph, removing pointless initializers
    model.clean_initializers()
    model.check_model()



#!/usr/bin/env python
# ******************************************************************************
# Copyright 2024 Brainchip Holdings Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ******************************************************************************

__all__ = ["convert_conv_to_gemm"]

import numpy as np
import onnx
import onnx.helper
import onnx.numpy_helper

from ...graph_tools import check_node_attributes, get_tensor_shape
from ..model import ONNXModel


def _find_conv_reshape_or_squeeze_patterns(model):
    # Finds and returns sublists of nodes in the model that match
    # (Conv, Reshape) or (Conv, Squeeze) patterns.
    node_patterns = []

    for node in model.nodes():
        node_outbounds = model.get_children(node)

        if node.op_type == "Conv" and len(node_outbounds) == 1:
            outbound = node_outbounds[0]
            if outbound.op_type in ("Reshape", "Squeeze"):
                node_patterns.append([node, outbound])

    return node_patterns


def _check_reshape_or_squeeze_output_shape_2D(model, squeeze_or_reshape_node):
    squeeze_or_reshape_output_shape = get_tensor_shape(
        model.find_value_info_by_name(squeeze_or_reshape_node.output[0]))

    if len(squeeze_or_reshape_output_shape) != 2:
        raise ValueError(
            f"Expected a 2D shape for the output of the node '{squeeze_or_reshape_node.name}', "
            f"but got a shape with {len(squeeze_or_reshape_output_shape)} dimensions: "
            f"{squeeze_or_reshape_output_shape}."
        )


def convert_conv_to_gemm(model):
    """
    Transforms Conv + (Reshape/Squeeze) into Flatten + Gemm if both operation chains are compatible.

    Args:
        model (ONNXModel): The ONNX model to be processed.
    """
    def _gemm_weight_from_conv(conv_node):
        # squeezes the conv weight to be compatible with dense weights
        # and updates the tensor proto
        conv_weight = model.get_variable(conv_node.input[1])
        conv_weight = np.reshape(conv_weight, (conv_weight.shape[0], -1))
        conv_weight_tp = model.get_initializer(conv_node.input[1])
        conv_weight_tp.CopyFrom(onnx.numpy_helper.from_array(conv_weight, conv_node.input[1]))

    def _update_conv_output_shape(conv_node):
        # Flatten node output will have the same channels as conv_input
        # but with conv output name
        conv_input_vi = model.find_value_info_by_name(conv_node.input[0])
        conv_output_vi = model.find_value_info_by_name(conv_node.output[0])
        channels, height, width = conv_input_vi.type.tensor_type.shape.dim[1:]

        conv_output_vi.type.tensor_type.shape.dim[1].dim_value = channels.dim_value * \
            height.dim_value * width.dim_value

        # Remove the last 2 dims of conv output
        conv_output_vi.type.tensor_type.shape.dim.pop()
        conv_output_vi.type.tensor_type.shape.dim.pop()

    assert isinstance(model, ONNXModel)

    nodes_to_add = []
    nodes_to_remove = []

    node_patterns = _find_conv_reshape_or_squeeze_patterns(model)

    if len(node_patterns) == 0:
        return

    supported_attributes = {'strides': [[1, 1]], 'group': [1]}

    for node_pattern in node_patterns:
        conv_node, squeeze_or_reshape_node = node_pattern

        conv_input_shape = get_tensor_shape(model.find_value_info_by_name(conv_node.input[0]))
        supported_attributes.update({'kernel_shape': [list(conv_input_shape[-2:])],
                                     'pads': [[0] * len(conv_input_shape)]})

        try:
            _check_reshape_or_squeeze_output_shape_2D(model, squeeze_or_reshape_node)
            check_node_attributes(conv_node, supported_attributes)

        except ValueError:
            # We don't change the model when the node attributes
            # do not match the required constraints
            continue

        _gemm_weight_from_conv(conv_node)
        _update_conv_output_shape(conv_node)

        flatten_node = onnx.helper.make_node("Flatten",
                                             inputs=[conv_node.input[0]],
                                             outputs=conv_node.output
                                             )
        gemm_node = onnx.helper.make_node("Gemm",
                                          inputs=[flatten_node.output[0]] + conv_node.input[1:],
                                          outputs=squeeze_or_reshape_node.output,
                                          transB=1
                                          )

        nodes_to_add.extend([flatten_node, gemm_node])
        nodes_to_remove.extend(node_pattern)

    model.add_nodes(nodes_to_add)
    model.remove_nodes(nodes_to_remove)
    model.clean_initializers()

    # As we add new nodes, we need to topologically sort the model graph
    model.topological_sort()
    model.check_model()



#!/usr/bin/env python
# ******************************************************************************
# Copyright 2023 Brainchip Holdings Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ******************************************************************************
__all__ = ["get_scale_out_ops"]

from onnx.helper import make_node


def get_scale_out_ops(in_name, out_name, scale_name="Scale", shift_name="Shift", saturate=True):
    """Return the scale out operation chain, following the steps:

    1. Apply shift and scale to inputs,
    2. Perform Round(x) as Floor(x + 0.5) (to workaround banker's rounding),
    3. (Optional) Clip in output range [-128, 127].

    Args:
        in_name (str): the input tensor name.
        out_name (str): the required output tensor name.
        scale_name (str, optional): the scale tensor name. Defaults to Scale.
        shift_name (str, optional): the shift tensor name. Defaults to Shift.
        saturate (bool, optional): whether to saturate the output. Defaults to True.

    Returns:
        list of NodeProto: the operation chain.
    """
    nodes = []
    # Apply shift + scale
    # Note: We apply first shift to avoid (on float) the saturation due to the mantissa.
    nodes.append(make_node("Div", [in_name, shift_name], [f"{in_name}/scaled"]))
    if scale_name:
        nodes[-1].output[0] = f"{in_name}/div"
        nodes.append(make_node("Mul", [f"{in_name}/div", scale_name], [f"{in_name}/scaled"]))
    # Round as Floor(x + 0.5)
    # Note: for positive shift, the results being integer, the rounding has no effect.
    # We therefore apply the same operations for both shifts.
    y_out_name = out_name if not saturate else f"{in_name}/q"
    nodes += [make_node("Constant", [], ["OneHalf"], value_float=0.5),
              make_node("Add", [f"{in_name}/scaled", "OneHalf"], [f"{in_name}/half"]),
              make_node("Floor", [f"{in_name}/half"], [y_out_name])]
    # 3. Clip in output range (optional)
    if saturate:
        nodes += [make_node("Constant", [], ["min_range"], value_float=-128.0),
                  make_node("Constant", [], ["max_range"], value_float=127.0),
                  make_node("Clip", [f"{in_name}/q", "min_range", "max_range"], [out_name])]
    return nodes



#!/usr/bin/env python
# ******************************************************************************
# Copyright 2022 Brainchip Holdings Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ******************************************************************************

__all__ = ["AlignedWeightQuantizer"]

import tensorflow as tf
import keras

from ...tensors import QFloat
from ..recorders import TensorRecorder, QFloatRecorder
from .quantizers import Quantizer


@keras.saving.register_keras_serializable()
class AlignedWeightQuantizer(Quantizer):
    """A uniform quantizer that converts a float Tensor to a QFloat representation.

    Unlike its sibling the WeightQuantizer, it does not evaluate the fractional bits and scales of
    the resulting QFloat, but instead aligns them on those of another QFloat input.

    Args:
        bitwidth (int, optional): the quantization bitwidth. Defaults to 8.
        signed (bool, optional): whether the quantizer expects signed values or unsigned.
            Defaults to True.
    """

    def __init__(self, bitwidth=8, signed=True, **kwargs):
        super().__init__(bitwidth, signed, **kwargs)
        # Add the object that will store the shift values.
        self.shift = TensorRecorder(self.name + "/shift")
        self.qweights = QFloatRecorder(self.name + "/qweights")

    def call(self, inputs, other):
        """Quantize the float inputs, aligned on another QFloat

        The quantization is done in several steps:

            1. Compute the quantization ranges,
            2. Evaluate the maximum fractional bits,
            3. Quantize the inputs as a QFloat,
            4. Align the QFloat fractional bits on the other.

        Args:
            inputs (tf.Tensor): the inputs tensor.
            other (:obj:`QFloat`): a tensor to align on.

        Returns:
            :obj:`QFloat`: a quantized tensor with the same scales and frac_bits as other.
        """
        if not isinstance(inputs, (tf.Tensor, tf.Variable)):
            raise ValueError(
                f"{type(inputs)} as first param is not supported."
                "AlignedWeightQuantizer only accepts tf.Tensor.")
        if not isinstance(other, QFloat):
            raise ValueError(
                f"{type(other)} as second param is not supported."
                "AlignedWeightQuantizer only accepts QFloat.")

        other_value_bits = other.fp.value_bits
        other_frac_bits = other.fp.frac_bits

        # Compute the symmetric quantization ranges from the inputs. When the bitwidth
        # is greater than 4 (e.g. 8 bits), experimentations have shown that models quantized
        # per axis (especially for the bias) have a better performance.
        # However for the 4bits, the performance was lower if we quantize per axis that's why
        # we keep it per tensor.
        if self.bitwidth > 4:
            # if the inputs are 0D or 1D vector, the ranges will have the same shape as the inputs.
            # In the other case for 2D or 3D (e.g. PositionEmbs), we reduce all axis
            # except the last.
            ndims = len(tf.shape(inputs))
            axis = list(range(ndims - 1))
        else:
            axis = None

        ranges = tf.math.reduce_max(tf.math.abs(inputs), axis=axis)
        # Evaluate the maximum fractional bits we can use for the specified ranges
        frac_bits = QFloat.max_frac_bits(self.value_bits, ranges, other.scales)
        # Remove the input-dependent fractional bits from the gradient tape to avoid a loopback
        frac_bits = tf.stop_gradient(frac_bits)
        # Clamp the ideal frac_bits to the other fractional bits, to avoid downscaling afterwards
        frac_bits = tf.minimum(frac_bits, other_frac_bits)
        # Quantize the inputs with the resulting fractional bits and other scales
        outputs = QFloat.quantize(inputs, self.value_bits, other.scales, frac_bits)
        # Now, upscale to the other fractional bits
        outputs, shift = outputs.upscale(other_frac_bits, other_value_bits)
        # record quantized weights (including shift and promote)
        self.qweights(outputs)
        # record shift values to be able to recreate and store unshifted weights
        self.shift(shift)
        return outputs



#!/usr/bin/env python
# ******************************************************************************
# Copyright 2024 Brainchip Holdings Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ******************************************************************************
__all__ = ["check_clip_relu_compatibility", "check_conv_depthwise_compatibility",
           "check_node_link_to_input", "check_node_has_one_initializer"]


from ..graph_tools import get_variable, get_field


def check_clip_relu_compatibility(clip_node, graph):
    """Checks the compatibility of a Clip node with ReLU activation in an ONNX graph.

    This function ensures that the Clip node has a minimum value input and that
    the min_value is set to zero, which is required for compatibility with ReLU.

    Args:
        clip_node (onnx.NodeProto): The Clip node to be checked.
        graph (onnx.GraphProto): The ONNX graph containing the Clip node.

    Raises:
        ValueError: If the Clip node does not have a min_value input.
        ValueError: If the min_value is not set to zero.
    """
    pre_msg_erro = "Impossible to handle Clip {} node: "

    if len(clip_node.input) == 1 or not clip_node.input[1]:
        raise ValueError(pre_msg_erro.format(clip_node.name) + "Expected a min_value input.")

    min_value = get_variable(clip_node.input[1], graph)
    if min_value != 0:
        raise ValueError(pre_msg_erro.format(clip_node.name) + "min_value must be zero.")


def check_conv_depthwise_compatibility(conv_node, graph):
    """Checks the compatibility of a convolutional node with depthwise convolution in an ONNX graph.

    This function ensures that the convolutional node can be transformed into a
    depthwise convolution. A convolution can be considered depthwise if the
    number of groups is equal to the number of filters and the kernel shape.
    Then, the resulting kernel shape will be (groups, 1, Kx, Ky).

    Args:
        conv_node (onnx.NodeProto): The convolutional node to be checked.
        graph (onnx.GraphProto): The ONNX graph containing the convolutional node.

    Raises:
        RuntimeError: If the groups attribute is equal to 1, indicating standard convolution.
        ValueError: If the number of groups does not equal the number of filters.
        ValueError: If the kernel shape in the input channel axis is not equal to 1.
    """
    groups = get_field(conv_node, "group", 1)

    # Retrieve kernel shape
    kernel_shape = get_variable(conv_node.input[1], graph).shape

    if groups == 1:
        raise RuntimeError("For groups = 1, it is handled by the Conv layer.")

    expect_kernel_shape = (groups, 1, *kernel_shape[2:])
    if kernel_shape[:2] != expect_kernel_shape[:2]:
        raise ValueError(f"Impossible to handle groups={groups} on {conv_node.name} node. "
                         f"A kernel format equal to {expect_kernel_shape} was expected "
                         f"(found {kernel_shape}).")


def check_node_link_to_input(node, graph):
    """Check the compatibility of a node to be converted into InputQuantizer:

    It must be connected to the input.

    Args:
        node (onnx.NodeProto): the node to be checked.
        graph (onnx.GraphProto): the ONNX graph containing the node.
    """
    ginputs = [x.name for x in graph.input]
    if node.input[0] not in ginputs:
        raise ValueError(f"Impossible to handle {node.name} ({node.op_type}): "
                         "it must be linked to one input.")


def check_node_has_one_initializer(node, graph):
    """Check the compatibility of a Mul/Add nodes to be converted into InputQuantizer:

    They must multiply/add a vector by a constant.

    Args:
        node (onnx.NodeProto): the node to be checked.
        graph (onnx.GraphProto): the ONNX graph containing the node.
    """
    initializers = [x.name for x in graph.initializer]
    if len(node.input) != 2 or node.input[1] not in initializers:
        raise ValueError(f"Impossible to handle {node.name} ({node.op_type}): "
                         "an initializer is expected in the second position.")



#!/usr/bin/env python
# ******************************************************************************
# Copyright 2023 Brainchip Holdings Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ******************************************************************************
__all__ = ['sanitize']

import os
import tempfile
import onnx
from pathlib import Path

from onnxruntime.quantization.shape_inference import quant_pre_process

from .convert_avg_to_gap import convert_avg_to_gap
from .convert_conv_to_gemm import convert_conv_to_gemm
from .convert_matmul_to_gemm import convert_matmul_to_gemm
from .convert_min_max_to_clip import convert_min_max_to_clip
from .fuse_relu_clip_nodes import fuse_relu_clip_nodes
from .insert_rescaling import insert_rescaling
from .remove_pad_nodes import fold_pad_into_conv
from .remove_reshape import remove_reshape
from .squeeze_reshape_to_flatten import convert_squeeze_reshape_to_flatten
from .untranspose_gemm_weights import untranspose_gemm_weights
from ..model import ONNXModel


def sanitize(model):
    """Sanitize a model preparing it for quantization.

    This is a wrapping successive calls to several model transformations
    which aims at making the model quantization ready.

    Args:
        model: the input model

    Returns:
        the sanitized model
    """
    assert isinstance(model, ONNXModel)

    # Clone model to prevent modification of the original one
    model = model.clone()

    # Replace operations to match with current ONNX version
    model.update_model_version()

    # Clean inputs/outputs
    model.clean_graph_io()

    # Perform optimization only if model is not quantized
    if not any(node.domain == "com.brainchip" for node in model.nodes()):
        with tempfile.TemporaryDirectory(prefix="pre.quant.") as quant_tmp_dir:
            # To perfom ONNXRuntime optimization, we would like to use
            # onnxruntime.quantization.quant_pre_process, to optimize the model (when required)
            # and infer the intermediate shapes.
            # However, it always expects to read the model from a path. That is why we
            # save the input model if it is not a path.
            tmp_model_path = os.path.join(quant_tmp_dir, "model.onnx")
            model.save_model_to_file(tmp_model_path)

            # We employ onnxruntime preprocessing to deduce shapes and implement optimization.
            # There exists an issue in onnxruntime related to the Constant Node
            # when it comes to shape computation using symbolic_shape_inference.
            # The problem arises when attempting to access the value of the constant
            # node attribute, but this node possesses other distinct attribute values
            # (refer to https://onnx.ai/onnx/operators/onnx__Constant.html#constant-13).
            # To resolve this, we bypass the symbolic shape inference and apply first
            # an optimization that folds these constants, and then apply by shape inference.
            quant_pre_process(tmp_model_path, tmp_model_path, skip_symbolic_shape=True)
            quant_pre_process(tmp_model_path, tmp_model_path)

            # Load the model
            model = ONNXModel(onnx.load_model(Path(tmp_model_path)))

    # Adds a Rescaling node to the model
    insert_rescaling(model)

    # Retranspose Gemm weights if they are transposed
    untranspose_gemm_weights(model)

    # Convert Squeeze/Reshape into Flatten when possible
    convert_squeeze_reshape_to_flatten(model)

    # Convert Min/Max into Clip
    convert_min_max_to_clip(model)

    # Remove pointless reshape nodes
    remove_reshape(model)

    # Fold pad into conv when possible
    fold_pad_into_conv(model)

    # Convert AveragePool to GAP when possible
    convert_avg_to_gap(model)

    # Convert Matmul to Gemm
    convert_matmul_to_gemm(model)

    # Convert Conv to Gemm when possible
    convert_conv_to_gemm(model)

    # Fuses ReLU/Clip nodes
    fuse_relu_clip_nodes(model)

    return model



from .activations import *
from .dense import *
from .dropout import *
from .multi_inbounds import *
from .normalization import *
from .quantizers import *
from .shiftmax import *
from .vit import *
from .depthwise_convolution import *
from .convolution import *
from .separable_convolution import *
from .reshaping import *
from .attention import *
from .pooling import *
from .recorders import *
from .rescaling import *
from .batch_normalization import *
from .quantization_params import *
from .output_observer import *
from .buffer_temp_conv import *
from .stateful_recurrent import *



#!/usr/bin/env python
# ******************************************************************************
# Copyright 2024 Brainchip Holdings Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ******************************************************************************

__all__ = ["convert_avg_to_gap"]

from ...graph_tools import get_tensor_shape, check_node_attributes
from ..model import ONNXModel


def convert_avg_to_gap(model):
    """
    Converts AveragePool node to GlobalAveragePool if the kernel shape is equal to
    spatial dimensions.

    Args:
        model (ONNXModel): The ONNX model to be processed.
    """
    assert isinstance(model, ONNXModel)

    supported_attributes = {'auto_pad': ['NOTSET']}

    for node in model.nodes():
        if node.op_type == 'AveragePool':
            value_info = model.find_value_info_by_name(node.input[0])

            # shape format : (B, C, H, W)
            input_shape = get_tensor_shape(value_info)

            try:
                supported_attributes.update({'kernel_shape': [list(input_shape[2:])],
                                             'pads': [[0] * len(input_shape)]})
                check_node_attributes(node, supported_attributes)
                node.op_type = "GlobalAveragePool"
                node.ClearField('attribute')

            except ValueError:
                # We continue when the node attributes
                # do not match the required constraints
                continue
    model.check_model()



#!/usr/bin/env python
# ******************************************************************************
# Copyright 2022 Brainchip Holdings Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ******************************************************************************
import tensorflow as tf
import keras

from ...debugging import assert_rank_at_least
from ..fixed_point import FixedPoint
from ..qfloat import QFloat


@tf.experimental.dispatch_for_api(tf.linalg.matmul)
def fp_fp_matmul(a: FixedPoint,
                 b: FixedPoint,
                 transpose_a=False,
                 transpose_b=False,
                 adjoint_a=False,
                 adjoint_b=False,
                 a_is_sparse=False,
                 b_is_sparse=False,
                 output_type=None,
                 name=None):
    """Multiplies matrix `a` by matrix `b`, producing `a` * `b`.

    Args:
        a (:obj:`FixedPoint`): a FixedPoint of rank > 1.
        b (:obj:`FixedPoint`): a FixedPoint with same rank as `a`.
        transpose_a (bool, optional): if `True`, `a` is transposed before multiplication.
            Defaults to False.
        transpose_b (bool, optional): if `True`, `b` is transposed before multiplication.
            Defaults to False.
        adjoint_a (bool, optional): if `True`, `a` is conjugated and transposed before
            multiplication. Defaults to False.
        adjoint_b (bool, optional): if `True`, `b` is conjugated and transposed before
            multiplication. Defaults to False.
        a_is_sparse (bool, optional): must be False, argument kept for compatibility with
            original tf.matmul. Defaults to False.
        b_is_sparse (bool, optional): must be False, argument kept for compatibility with
            original tf.matmul. Defaults to False.
        output_type (NoneType, optional): must be None, argument kept for compatibility
            with original tf.matmul. Defaults to None.
        name (str, optional): the name for the operation. Defaults to None.

    Returns:
        :obj:`FixedPoint`: the multiplied FixedPoint.
    """
    if a_is_sparse:
        raise ValueError(
            f"Unsupported argument: a_is_sparse, provided {a_is_sparse}")
    if b_is_sparse:
        raise ValueError(
            f"Unsupported argument: b_is_sparse, provided {b_is_sparse}")
    if output_type is not None:
        raise ValueError(
            f"Unsupported argument: output_type, provided {output_type}")

    # We don't support matmul between vectors
    assert_rank_at_least(a.values, 2)
    assert_rank_at_least(b.values, 2)

    # Since the last dimension is collapsed by the matmul, (a,b) must be quantized per-tensor
    if not transpose_a:
        a.assert_per_tensor()
    if transpose_b:
        b.assert_per_tensor()

    # Do matmul on values
    m_values = tf.matmul(a.values, b.values, transpose_a, transpose_b, adjoint_a, adjoint_b, name)

    # Return a new FixedPoint
    return FixedPoint(m_values, a.value_bits, a.frac_bits + b.frac_bits)


@tf.experimental.dispatch_for_api(tf.linalg.matmul)
def fp_qf_matmul(a: FixedPoint,
                 b: QFloat,
                 transpose_a=False,
                 transpose_b=False,
                 adjoint_a=False,
                 adjoint_b=False,
                 a_is_sparse=False,
                 b_is_sparse=False,
                 output_type=None,
                 name=None):
    """Multiplies matrix `a` by matrix `b`, producing `a` * `b`.

    Args:
        a (:obj:`FixedPoint`): a FixedPoint of rank > 1.
        b (:obj:`QFloat`): a QFloat with same rank as `a`.
        transpose_a (bool, optional): if `True`, `a` is transposed before multiplication.
            Defaults to False.
        transpose_b (bool, optional): if `True`, `b` is transposed before multiplication.
            Defaults to False.
        adjoint_a (bool, optional): if `True`, `a` is conjugated and transposed before
            multiplication. Defaults to False.
        adjoint_b (bool, optional): if `True`, `b` is conjugated and transposed before
            multiplication. Defaults to False.
        a_is_sparse (bool, optional): must be False, argument kept for compatibility with
            original tf.matmul. Defaults to False.
        b_is_sparse (bool, optional): must be False, argument kept for compatibility with
            original tf.matmul. Defaults to False.
        output_type (NoneType, optional): must be None, argument kept for compatibility
            with original tf.matmul. Defaults to None.
        name (str, optional): the name for the operation. Defaults to None.

    Returns:
        :obj:`QFloat`: the multiplied QFloat.
    """
    # If b is transposed, its last dimension will collapse, so it must be per-tensor
    if transpose_b:
        b.assert_per_tensor()

    # Do matmul on a and b inner FixedPoint
    fp_prod = tf.matmul(a, b.fp, transpose_a, transpose_b, adjoint_a, adjoint_b, name)

    # Return a new QFloat with the same scales
    return QFloat(fp_prod, b.scales)


@tf.experimental.dispatch_for_api(tf.linalg.matmul)
def qf_qf_matmul(a: QFloat,
                 b: QFloat,
                 transpose_a=False,
                 transpose_b=False,
                 adjoint_a=False,
                 adjoint_b=False,
                 a_is_sparse=False,
                 b_is_sparse=False,
                 output_type=None,
                 name=None):
    """Multiplies matrix `a` by matrix `b`, producing `a` * `b`.

    Args:
        a (:obj:`QFloat`): a QFloat of rank > 1.
        b (:obj:`QFloat`): a QFloat with same rank as `a`.
        transpose_a (bool, optional): if `True`, `a` is transposed before multiplication.
            Defaults to False.
        transpose_b (bool, optional): if `True`, `b` is transposed before multiplication.
            Defaults to False.
        adjoint_a (bool, optional): if `True`, `a` is conjugated and transposed before
            multiplication. Defaults to False.
        adjoint_b (bool, optional): if `True`, `b` is conjugated and transposed before
            multiplication. Defaults to False.
        a_is_sparse (bool, optional): must be False, argument kept for compatibility with
            original tf.matmul. Defaults to False.
        b_is_sparse (bool, optional): must be False, argument kept for compatibility with
            original tf.matmul. Defaults to False.
        output_type (NoneType, optional): must be None, argument kept for compatibility
            with original tf.matmul. Defaults to None.
        name (str, optional): the name for the operation. Defaults to None.

    Returns:
        :obj:`QFloat`: the multiplied QFloat.
    """
    # the first input must be per-tensor to be able to multiply the scales
    a.assert_per_tensor()
    # The second input must also be per-tensor if it is transposed
    if transpose_b:
        b.assert_per_tensor()

    # Do matmul on a and b inner FixedPoint
    fp_prod = tf.matmul(a.fp, b.fp, transpose_a, transpose_b, adjoint_a, adjoint_b, name)

    # Return a new QFloat with multiplied scales
    return QFloat(fp_prod, a.scales * b.scales)


@tf.experimental.dispatch_for_api(keras.backend.depthwise_conv2d)
def fp_fp_depthwise_conv2d(x: FixedPoint,
                           depthwise_kernel: FixedPoint,
                           strides=(1, 1),
                           padding='valid',
                           data_format=None,
                           dilation_rate=(1, 1)):
    """ 2D convolution with separable filters.

    Args:
        x (obj:`FixedPoint`): input tensor.
        depthwise_kernel (:obj:`FixedPoint`): convolution kernel for the depthwise convolution.
        strides (tuple, optional): strides tuple (length 2). Defaults to (1, 1).
        padding (str, optional): `"same"` or `"valid"`. Defaults to 'valid'.
        data_format (str, optional): `"channels_last"` or `"channels_first"`. Defaults to None.
        dilation_rate (tuple, optional): tuple of integers, dilation rates for the separable
            convolution. Defaults to (1, 1).

    Returns:
        :obj:`FixedPoint`: output tensor.

    """
    # Unlike other convolutions, the depthwise does not require its inputs to
    # be quantized per-tensor as the input channels are processed independently

    # Do convolution on values
    conv_values = keras.backend.depthwise_conv2d(x.values, depthwise_kernel.values, strides,
                                                 padding, data_format, dilation_rate)

    # Evaluate the outputs fractional bits
    kernel_frac_bits = depthwise_kernel.frac_bits
    if not depthwise_kernel.per_tensor:
        # Remove the last dimension of the broadcastable depthwise filters frac bits
        kernel_frac_bits = tf.squeeze(kernel_frac_bits)
    # The output fractional bits are the sum of inputs and filters frac bits
    frac_bits = x.frac_bits + kernel_frac_bits

    # Return a new FixedPoint
    return FixedPoint(conv_values, x.value_bits, frac_bits)


@tf.experimental.dispatch_for_api(keras.backend.depthwise_conv2d)
def fp_qf_depthwise_conv2d(x: FixedPoint,
                           depthwise_kernel: QFloat,
                           strides=(1, 1),
                           padding='valid',
                           data_format=None,
                           dilation_rate=(1, 1)):
    """ 2D convolution with separable filters.

    Args:
        x (obj:`FixedPoint`): input tensor.
        depthwise_kernel (:obj:`QFloat`): convolution kernel for the depthwise convolution.
        strides (tuple, optional): strides tuple (length 2). Defaults to (1, 1).
        padding (str, optional): `"same"` or `"valid"`. Defaults to 'valid'.
        data_format (str, optional): `"channels_last"` or `"channels_first"`. Defaults to None.
        dilation_rate (tuple, optional): tuple of integers, dilation rates for the separable
            convolution. Defaults to (1, 1).

    Returns:
        :obj:`QFloat`: output tensor.

    """
    # Do convolution on x and depthwise_kernel inner FixedPoint
    fp_conv = keras.backend.depthwise_conv2d(x, depthwise_kernel.fp, strides,
                                             padding, data_format, dilation_rate)

    # Evaluate the output scales
    scales = depthwise_kernel.scales
    if not depthwise_kernel.per_tensor:
        # Remove the last dimension of the broadcastable depthwise filters scales
        scales = tf.squeeze(scales)
    # Return a new QFloat with the same scales
    return QFloat(fp_conv, scales)


@tf.experimental.dispatch_for_api(tf.nn.convolution)
def fp_fp_convolution(
        input: FixedPoint,
        filters: FixedPoint,
        strides=None, padding="VALID", data_format=None, dilations=None, name=None):
    """Perform the convolution operation between input and filter tensors.

    Args:
        input (:obj:`FixedPoint`): The input FixedPoint.
        filters (:obj:`QFloat`): The filters FixedPoint.
        strides (list, optional): Sequence of N ints >= 1. Specifies the output stride.
            Defaults to None.
        padding (str, optional): A string, either `"VALID"` or `"SAME"`. The padding algorithm.
            Defaults to "VALID"
        data_format (str, optional): Specifies whether the channel dimension of
            the `input` and output is the last dimension (default, or if `data_format`
            does not start with "NC"), or the second dimension (if `data_format`
            starts with "NC").  For N=1, the valid values are "NWC" (default) and
            "NCW".  For N=2, the valid values are "NHWC" (default) and "NCHW".
            For N=3, the valid values are "NDHWC" (default) and "NCDHW".
        dilations (list, optional): Alias for dilation_rate. Sequence of N ints >= 1.
            Specifies the filter upsampling/input downsampling rate. Defaults to None.
        name (str, optional): the name for the Tensorflow ops. Defaults to None.

    Returns:
        :obj:`FixedPoint`: a FixedPoint containing the output values.
    """
    # Input must be quantized per-tensor because the products are eventually summed
    input.assert_per_tensor()

    # Do convolution on values
    conv_values = tf.nn.convolution(input.values, filters.values, strides, padding, data_format,
                                    dilations, name)

    # Return a new FixedPoint
    return FixedPoint(conv_values, input.value_bits, input.frac_bits + filters.frac_bits)


@tf.experimental.dispatch_for_api(tf.nn.convolution)
def fp_qf_convolution(
        input: FixedPoint,
        filters: QFloat,
        strides=None, padding="VALID", data_format=None, dilations=None, name=None):
    """Perform the convolution operation between input and filter tensors.

    Args:
        input (:obj:`FixedPoint`): The input FixedPoint.
        filters (:obj:`QFloat`): The filters QFloat.
        strides (list, optional): Sequence of N ints >= 1. Specifies the output stride.
            Defaults to None.
        padding (str, optional): A string, either `"VALID"` or `"SAME"`. The padding algorithm.
            Defaults to "VALID"
        data_format (str, optional): Specifies whether the channel dimension of
            the `input` and output is the last dimension (default, or if `data_format`
            does not start with "NC"), or the second dimension (if `data_format`
            starts with "NC"). For N=1, the valid values are "NWC" (default) and
            "NCW".  For N=2, the valid values are "NHWC" (default) and "NCHW".
            For N=3, the valid values are "NDHWC" (default) and "NCDHW".
        dilations (list, optional): Alias for dilation_rate. Sequence of N ints >= 1.
            Specifies the filter upsampling/input downsampling rate. Defaults to None.
        name (str, optional): the name for the Tensorflow ops. Defaults to None.

    Returns:
        :obj:`QFloat`: a QFloat containing the output values.
    """
    # Do convolution on input and filters inner FixedPoint
    fp_conv = tf.nn.convolution(input, filters.fp, strides, padding, data_format,
                                dilations, name)

    # Return a new QFloat with the same scales
    return QFloat(fp_conv, filters.scales)


@tf.experimental.dispatch_for_api(tf.nn.convolution)
def qf_qf_convolution(
        input: QFloat,
        filters: QFloat,
        strides=None, padding="VALID", data_format=None, dilations=None, name=None):
    """Perform the convolution operation between input and filter tensors.

    Args:
        input (:obj:`QFloat`): The input QFloat.
        filters (:obj:`QFloat`): The filters QFloat.
        strides (list, optional): Sequence of N ints >= 1. Specifies the output stride.
            Defaults to None.
        padding (str, optional): A string, either `"VALID"` or `"SAME"`. The padding algorithm.
            Defaults to "VALID"
        data_format (str, optional): Specifies whether the channel dimension of
            the `input` and output is the last dimension (default, or if `data_format`
            does not start with "NC"), or the second dimension (if `data_format`
            starts with "NC"). For N=1, the valid values are "NWC" (default) and
            "NCW". For N=2, the valid values are "NHWC" (default) and "NCHW".
            For N=3, the valid values are "NDHWC" (default) and "NCDHW".
        dilations (list, optional): Alias for dilation_rate. Sequence of N ints >= 1.
            Specifies the filter upsampling/input downsampling rate. Defaults to None.
        name (str, optional): the name for the Tensorflow ops. Defaults to None.

    Returns:
        :obj:`QFloat`: a QFloat containing the output values.
    """
    # Input must be quantized per-tensor because the scales need to be multiplied
    input.assert_per_tensor()

    # Do convolution on input and filters inner FixedPoint(s)
    conv_fp = tf.nn.convolution(input.fp, filters.fp, strides, padding, data_format,
                                dilations, name)

    # Evaluate the resulting scales as the product of both scales
    conv_scales = input.scales * filters.scales
    return QFloat(conv_fp, conv_scales)


@tf.experimental.dispatch_for_api(keras.backend.conv2d_transpose)
def fp_fp_conv2d_transpose(
        x: FixedPoint,
        kernel: FixedPoint,
        output_shape, strides=(1, 1), padding='valid', data_format=None, dilation_rate=(1, 1)):
    """The transpose of conv2d.

    This operation is sometimes called "deconvolution" after (Zeiler et al., 2010), but is
    really the transpose (gradient) of atrous_conv2d rather than an actual deconvolution.

    Args:
        x (:obj:`FixedPoint`): A 4-D FixedPoint of shape [batch, height, width, in_channels] for
            NHWC data format or [batch, in_channels, height, width] for NCHW data format.
        kernel (:obj:`FixedPoint`): A FixedPoint with a [height, width, out_channels, in_channels]
            shape. Kernel's in_channels dimension must match that of input.
        output_shape (tf.Tensor): A 1-D Tensor representing the output shape of the
            deconvolution op.
        strides (list, optional): Sequence of 2 ints >= 1. Specifies the output stride.
            Defaults to (1, 1).
        padding (str, optional): A string, either `'valid'` or `'same'`. The padding algorithm.
            Defaults to 'valid'
        data_format (str, optional): Specifies whether the channel dimension of
            the `input` and output is the last dimension (default, or if `data_format`
            does not start with "NC"), or the second dimension (if `data_format`
            starts with "NC"). For N=1, the valid values are "NWC" (default) and
            "NCW". For N=2, the valid values are "NHWC" (default) and "NCHW".
            For N=3, the valid values are "NDHWC" (default) and "NCDHW".
        dilation_rate (list, optional): Sequence of 2 ints >= 1.
            Specifies the filter upsampling/input downsampling rate. Defaults to (1, 1).

    Returns:
        :obj:`FixedPoint`: a FixedPoint containing the output values.
    """
    # Input must be quantized per-tensor because the products are eventually summed
    x.assert_per_tensor()

    # Transpose filter values so that they have the shape expected by the tensorflow op:
    #  (h, w, filters, channels)
    filter_values = tf.transpose(kernel.values, (0, 1, 3, 2))

    # Do conv2d_transpose on values
    conv_values = keras.backend.conv2d_transpose(x.values, filter_values, output_shape, strides,
                                                 padding, data_format, dilation_rate)

    # Return a new FixedPoint
    return FixedPoint(conv_values, x.value_bits, x.frac_bits + kernel.frac_bits)


@tf.experimental.dispatch_for_api(keras.backend.conv2d_transpose)
def fp_qf_conv2d_transpose(
        x: FixedPoint,
        kernel: QFloat,
        output_shape, strides=(1, 1), padding='valid', data_format=None, dilation_rate=(1, 1)):
    """The transpose of conv2d.

    This operation is sometimes called "deconvolution" after (Zeiler et al., 2010), but is
    really the transpose (gradient) of atrous_conv2d rather than an actual deconvolution.

    Args:
        x (:obj:`FixedPoint`): A 4-D FixedPoint of shape [batch, height, width, in_channels] for
            NHWC data format or [batch, in_channels, height, width] for NCHW data format.
        kernel (:obj:`QFloat`): A QFloat of shape [height, width, out_channels, in_channels]
            shape. Kernel's in_channels dimension must match that of input.
        output_shape (tf.Tensor): A 1-D Tensor representing the output shape of the
            deconvolution op.
        strides (list, optional): Sequence of 2 ints >= 1. Specifies the output stride.
            Defaults to (1, 1).
        padding (str, optional): A string, either `'valid'` or `'same'`. The padding algorithm.
            Defaults to 'valid'
        data_format (str, optional): Specifies whether the channel dimension of
            the `input` and output is the last dimension (default, or if `data_format`
            does not start with "NC"), or the second dimension (if `data_format`
            starts with "NC"). For N=1, the valid values are "NWC" (default) and
            "NCW". For N=2, the valid values are "NHWC" (default) and "NCHW".
            For N=3, the valid values are "NDHWC" (default) and "NCDHW".
        dilation_rate (list, optional): Sequence of 2 ints >= 1.
            Specifies the filter upsampling/input downsampling rate. Defaults to (1, 1).

    Returns:
        :obj:`QFloat`: a QFloat containing the output values.
    """
    # Input must be quantized per-tensor because the products are eventually summed
    x.assert_per_tensor()

    # Do conv2d_transpose on inputs and filter inner FixedPoint
    conv_values = keras.backend.conv2d_transpose(x, kernel.fp, output_shape, strides,
                                                 padding, data_format, dilation_rate)

    # Return a new QFloat
    return QFloat(conv_values, kernel.scales)


@tf.experimental.dispatch_for_api(tf.nn.max_pool)
def fp_max_pool2d(
        input: FixedPoint,
        ksize, strides, padding, data_format=None, name=None):
    """Perform the max pooling operation

    Downsamples the input along its spatial dimensions by taking the maximum value over an input
    window for each channel of the input. The window is shifted by strides along each dimension.

    Args:
        input (:obj:`FixedPoint`): A 4-D FixedPoint of shape [batch, height, width, in_channels] for
            NHWC data format or [batch, in_channels, height, width] for NCHW data format.
        ksize (list): An int or list of ints that has length 1, N or N+2. The size of the window
            for each dimension of the input tensor.
        strides (list): Sequence of N ints >= 1. Specifies the output stride.
        padding (str, optional): A string, either `"VALID"` or `"SAME"`. The padding algorithm.
            Defaults to "VALID"
        data_format (str, optional): Specifies the channel dimension. For N=1 it can be
            either "NWC" (default) or "NCW", for N=2 it can be either "NHWC" (default) or "NCHW" and
            for N=3 either "NDHWC" (default) or "NCDHW".
        name (str, optional): the name for the Tensorflow ops. Defaults to None.

    Returns:
        :obj:`FixedPoint`: a FixedPoint containing the output values.
    """
    # Do max_pool on values
    max_pool_values = tf.nn.max_pool(input.values, ksize, strides, padding, data_format, name)

    # Return a new FixedPoint
    return FixedPoint(max_pool_values, input.value_bits, input.frac_bits)


@tf.experimental.dispatch_for_api(tf.nn.max_pool)
def qf_max_pool2d(
        input: QFloat,
        ksize, strides, padding, data_format=None, name=None):
    """Perform the max pooling operation

    Downsamples the input along its spatial dimensions by taking the maximum value over an input
    window for each channel of the input. The window is shifted by strides along each dimension.

    Args:
        input (:obj:`QFloat`): A 4-D QFloat of shape [batch, height, width, in_channels] for
            NHWC data format or [batch, in_channels, height, width] for NCHW data format.
        ksize (int, list): An int or list of ints that has length 1, N or N+2. The size of the
            window for each dimension of the input tensor.
        strides (list): Sequence of N ints >= 1. Specifies the output stride.
        padding (str, optional): A string, either `"VALID"` or `"SAME"`. The padding algorithm.
            Defaults to "VALID"
        data_format (str, optional): Specifies the channel dimension. For N=1 it can be
            either "NWC" (default) or "NCW", for N=2 it can be either "NHWC" (default) or "NCHW" and
            for N=3 either "NDHWC" (default) or "NCDHW".
        name (str, optional): the name for the Tensorflow ops. Defaults to None.

    Returns:
        :obj:`QFloat`: a QFloat containing the output values.
    """
    # Do max_pool on inputs inner FixedPoint
    max_pool_fp = tf.nn.max_pool(input.fp, ksize, strides, padding, data_format, name)

    # Return a new QFloat
    return QFloat(max_pool_fp, input.scales)



#!/usr/bin/env python
# ******************************************************************************
# Copyright 2024 Brainchip Holdings Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ******************************************************************************
__all__ = ['ONNXModel']

import warnings

import onnx
import onnx.numpy_helper
import onnx.version_converter
from onnxruntime.quantization.onnx_model import ONNXModel as Model

from ..graph_tools import get_tensor_shape, get_tensor_dtype, find_value_info_by_name


class ONNXModel(Model):
    """Interface with some tools to handle ``onnx.ModelProto`` objects

    Args:
        Model (``onnx.ModelProto`): the base model
    """

    def __init__(self, model):
        super().__init__(model)
        self.check_model()

    @property
    def name(self):
        return self.model.graph.name

    @property
    def input(self):
        return self.model.graph.input

    @property
    def output(self):
        return self.model.graph.output

    @property
    def op_type_nodes(self):
        return tuple(x.op_type for x in self.nodes())

    @property
    def serialized(self):
        return self.model.SerializeToString()

    def check_model(self):
        """Check the consistency of a model.

        An exception is raised if the test fails.
        """
        onnx.checker.check_model(self.model, full_check=True)

    def clean_graph_io(self):
        # Remove the 'inputs' and 'outputs' that are contained in initializer graph field:
        # these are constants and may not be considered as inputs/outputs of the graph
        for value_info in self.input[:]:
            initializer = self.get_initializer(value_info.name)
            if initializer is not None:
                self.input.remove(value_info)

        for value_info in self.output[:]:
            initializer = self.get_initializer(value_info.name)
            if initializer is not None:
                self.output.remove(value_info)

    def update_model_version(self):
        # Try to replace the ONNX version in the graph with the current one
        version = onnx.defs.onnx_opset_version()
        try:
            self.model = onnx.version_converter.convert_version(self.model, target_version=version)
        except Exception as e:
            warnings.warn(f"Impossible to convert model in version {version}. The model may not be "
                          f"compatible with the quantization pipeline. Reason: \n{str(e)}")

        # Once the opset has been updated, IR_VERSION should be compatible with the latest version.
        self.model.ir_version = onnx.IR_VERSION

    def clone(self):
        model = onnx.ModelProto()
        model.CopyFrom(self.model)
        return ONNXModel(model)

    def get_variable(self, name):
        """Helper to get the value of an initializar as np.array.

        Args:
            name (str): the name of the variable.

        Returns:
            np.array: the value of the variable.
        """
        initializer = self.get_initializer(name)
        assert initializer, f"'{name}' was not found in initializer field."
        return onnx.numpy_helper.to_array(initializer)

    def get_input_shape(self, input_name=None):
        """Read the input shape(s) in the graph.

        Args:
            tensor_name (str, optional): return the shape only for this tensor. Defaults to None.

        Returns:
            dict or tuple: the shape of each input in the graph
        """
        if input_name is not None:
            input_value_info = self.find_value_info_by_name(input_name)
            assert input_value_info, f"{input_name} does not exist in the graph inputs."
            return get_tensor_shape(input_value_info)
        return {x.name: get_tensor_shape(x) for x in self.input}

    def get_input_dtype(self, input_name=None):
        """Read the input type(s) in the graph.

        Args:
            tensor_name (str, optional): return the type only for this tensor. Defaults to None.

        Returns:
            dict or np.ndarray: the type of each input in the graph
        """
        if input_name is not None:
            input_value_info = self.find_value_info_by_name(input_name)
            assert input_value_info, f"{input_name} does not exist in the graph inputs."
            return get_tensor_dtype(input_value_info)
        return {x.name: get_tensor_dtype(x) for x in self.input}

    def find_node_by_name(self, node_name):
        """Find a node by its name in the graph

        Args:
            node_name (str): the node name

        Returns:
            NodeProto: the node found
        """
        return super().find_node_by_name(node_name, [], self.graph())

    def find_value_info_by_name(self, tensor_name):
        """Return a value info by its name

        Args:
            tensor_name (str): the tensor name

        Returns:
            NodeProto: the node found
        """
        return find_value_info_by_name(self.graph(), tensor_name)

    def get_node_inputs(self, node):
        """Return the set of non initializer inputs in a node

        Args:
            node (NodeProto): the node to extract the inputs

        Returns:
            list: the non initializer inputs
        """
        initializer_names = self.get_initializer_name_set()
        non_initializer_inputs = []
        for input in node.input:
            # check input is not an initializer and it's not empty (e.g. '')
            if input not in initializer_names and len(input) > 0:
                non_initializer_inputs.append(input)
        return non_initializer_inputs

    def _clear_node_initializers(self, node):
        initializers_to_remove = []

        for input_name in node.input:
            initializer = self.get_initializer(input_name)
            if initializer and len(self.find_nodes_by_initializer(self.graph(), initializer)) == 1:
                initializers_to_remove.append(initializer)

        self.remove_initializers(initializers_to_remove)

    def _remove_node_metadata(self, node_to_remove):
        # Case for skip connections
        non_initializer_inputs = self.get_node_inputs(node_to_remove)

        if len(non_initializer_inputs) > 1:
            raise ValueError(
                f"Cannot remove node {node_to_remove.name} ({node_to_remove.op_type}) as it has "
                f"more than one non-initializer input."
            )

        target_value_info = self.find_value_info_by_name(node_to_remove.output[0])

        # Case where node_to_remove is the last node
        if target_value_info in self.output:
            # Rename graph output name. In this case, the value info is the graph output
            target_value_info.name = node_to_remove.input[0]

        else:
            # Update the input connections of subsequent nodes to bypass node_to_remove
            self.replace_input_of_all_nodes(node_to_remove.output[0], node_to_remove.input[0])
            self._clear_node_initializers(node_to_remove)

            # Remove node_to_remove output value info
            if target_value_info is not None:
                self.graph().value_info.remove(target_value_info)

    def remove_node(self, node_to_remove, update_graph=False):
        """
        Removes a specified node from the ONNX model's graph and optionally updates the graph.

        Args:
            node_to_remove (NodeProto): The node object to be removed from the graph.
            update_graph (bool, optional): If True, the graph's structure will be updated to handle
                the removal by adjusting the outbound inputs. Defaults to False.
        """
        if update_graph:
            self._remove_node_metadata(node_to_remove)
        super().remove_node(node_to_remove)

    def remove_nodes(self, nodes_to_remove, update_graph=False):
        """
        Removes a list of specified nodes from the ONNX model's graph
        and optionally updates the graph.

        Args:
            nodes_to_remove (list): A list of node objects to be removed from the graph.
            update_graph (bool, optional): If True, the graph's structure will be updated to handle
                the removal by adjusting the outbound inputs. Defaults to False.
        """
        for node in nodes_to_remove:
            self.remove_node(node, update_graph)



#!/usr/bin/env python
# ******************************************************************************
# Copyright 2023 Brainchip Holdings Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ******************************************************************************
import os
import tensorflow as tf


def assert_enabled():
    """Check if run-time assertions are enabled

    Assertions are enabled by default. They can be disabled by setting the
    "ASSERT_ENABLED" environment variable to "0".

    Returns:
        bool: True if assertions are enabled.
    """
    return os.environ.get("ASSERT_ENABLED", "1") == "1"


def assert_equal(x, y, message=None):
    """Assert the condition `x == y` holds element-wise.

    If assertions are enabled, this wraps the corresponding tensorflow assertion.
    Otherwise it does nothing.

    Args:
      x (tf.Tensor): a tensor
      y (tf.Tensor): a tensor of the same dtype as and broadcastable to `x`.
      message (str): A string to prefix to the default message. Defaults to None.

    Raises:
      InvalidArgumentError: if the check can be performed immediately and is False.
      The check can be performed immediately in eager mode or if `x` and `y` are statically known.
    """
    if assert_enabled():
        tf.debugging.assert_equal(x, y, message)


def assert_none_equal(x, y, message=None):
    """Assert the condition `x != y` holds element-wise.

    If assertions are enabled, this wraps the corresponding tensorflow assertion.
    Otherwise it does nothing.

    Args:
      x (tf.Tensor): a tensor
      y (tf.Tensor): a tensor of the same dtype as and broadcastable to `x`.
      message (str): A string to prefix to the default message. Defaults to None.

    Raises:
      InvalidArgumentError: if the check can be performed immediately and is False.
      The check can be performed immediately in eager mode or if `x` and `y` are statically known.
    """
    if assert_enabled():
        tf.debugging.assert_none_equal(x, y, message=message)


def assert_less(x, y, message=None):
    """Assert the condition `x < y` holds element-wise.

    If assertions are enabled, this wraps the corresponding tensorflow assertion.
    Otherwise it does nothing.

    Args:
      x (tf.Tensor): a tensor
      y (tf.Tensor): a tensor broadcastable to `x`.
      message (str): A string to prefix to the default message. Defaults to None.

    Raises:
      InvalidArgumentError: if the check can be performed immediately and is False.
      The check can be performed immediately in eager mode or if `x` and `y` are statically known.
    """
    if assert_enabled():
        tf.debugging.assert_less(tf.cast(x, tf.float32), tf.cast(y, tf.float32), message)


def assert_less_equal(x, y, message=None):
    """Assert the condition `x <= y` holds element-wise.

    If assertions are enabled, this wraps the corresponding tensorflow assertion.
    Otherwise it does nothing.

    Args:
      x (tf.Tensor): a tensor
      y (tf.Tensor): a tensor broadcastable to `x`.
      message (str): A string to prefix to the default message. Defaults to None.

    Raises:
      InvalidArgumentError: if the check can be performed immediately and is False.
      The check can be performed immediately in eager mode or if `x` and `y` are statically known.
    """
    if assert_enabled():
        tf.debugging.assert_less_equal(tf.cast(x, tf.float32), tf.cast(y, tf.float32), message)


def assert_rank_at_least(x, rank, message=None):
    """Assert that `x` has rank of at least `rank`.

    If assertions are enabled, this wraps the corresponding tensorflow assertion.
    Otherwise it does nothing.

    Args:
      x (tf.Tensor): a tensor
      rank (tf.Tensor): a scalar integer tensor
      message (str): A string to prefix to the default message. Defaults to None.

    Raises:
      InvalidArgumentError: if the check cannot be performed statically and is False.
      ValueError: if the check can be performed statically and is False.
    """
    if assert_enabled():
        tf.debugging.assert_rank_at_least(x, rank, message)



#!/usr/bin/env python
# ******************************************************************************
# Copyright 2022 Brainchip Holdings Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ******************************************************************************

__all__ = ["ClassToken", "QuantizedClassToken",
           "AddPositionEmbs", "QuantizedAddPositionEmbs",
           "ExtractToken", "QuantizedExtractToken"]

import keras
import tensorflow as tf

from .layers_base import (register_quantize_target, register_no_output_quantizer, rescale_outputs,
                          tensor_inputs, register_aligned_inputs, apply_buffer_bitwidth,
                          QuantizedLayer)
from .quantizers import AlignedWeightQuantizer, OutputQuantizer
from ..tensors import QFloat, FixedPoint


@keras.saving.register_keras_serializable()
class ClassToken(keras.layers.Layer):
    """Append a class token to an input layer.

    Args:
        initializer (keras.initializers.Initializer): Initializer for the class
            variable. Defaults to None.
    """

    def __init__(self, *args, initializer=None, **kwargs):
        super().__init__(*args, **kwargs)
        if initializer is None:
            initializer = keras.initializers.TruncatedNormal(stddev=0.02)
        self.initializer = initializer

    def build(self, input_shape):
        # Ensure variables are build with the appropriate name
        with tf.name_scope(self.name + '/'):
            super().build(input_shape)
            self.hidden_size = input_shape[-1]
            self.cls = tf.Variable(
                name="cls",
                initial_value=self.initializer(
                    shape=(1, 1, self.hidden_size), dtype="float32"),
                trainable=True,
            )

    def call(self, inputs):
        batch_size = tf.shape(inputs)[0]
        cls_broadcasted = tf.cast(
            tf.broadcast_to(self.cls, [batch_size, 1, self.hidden_size]),
            dtype=inputs.dtype,
        )
        return tf.concat([cls_broadcasted, inputs], 1)


@keras.saving.register_keras_serializable()
class AddPositionEmbs(keras.layers.Layer):
    """Adds (optionally learned) positional embeddings to the inputs.

    Args:
        initializer (keras.initializers.Initializer): Initializer for the class
            variable. Defaults to None.
    """

    def __init__(self, *args, initializer=None, **kwargs):
        super().__init__(*args, **kwargs)
        if initializer is None:
            initializer = keras.initializers.TruncatedNormal(stddev=0.02)
        self.initializer = initializer

    def build(self, input_shape):
        assert len(
            input_shape) == 3, f"Number of dimensions should be 3, got {len(input_shape)}"
        # Ensure variables are build with the appropriate name
        with tf.name_scope(self.name + '/'):
            super().build(input_shape)
            self.pe = tf.Variable(
                name="pos_embedding",
                initial_value=self.initializer(
                    shape=(1, input_shape[1], input_shape[2])),
                dtype="float32",
                trainable=True,
            )

    def call(self, inputs):
        return inputs + tf.cast(self.pe, dtype=inputs.dtype)


@register_quantize_target(ClassToken)
@register_no_output_quantizer
@keras.saving.register_keras_serializable()
class QuantizedClassToken(QuantizedLayer, ClassToken):
    """Quantize the :class:`ClassToken` layer, allowing quantization of the output.

    Args:
        quant_config (dict, optional): the serialized quantization configuration. Defaults to None.
    """

    def __init__(self, *args, quant_config=None, **kwargs):
        super().__init__(*args, quant_config=quant_config, **kwargs)
        cls_quantizer_cfg = self.quant_config.get("cls_quantizer", {})
        self.cls_quantizer = AlignedWeightQuantizer(name="cls_quantizer", **cls_quantizer_cfg)
        self.buffer_bitwidth = apply_buffer_bitwidth(self.quant_config, signed=True)

    @tensor_inputs([QFloat])
    def call(self, inputs):
        batch_size = tf.shape(inputs.values)[0]

        # Quantize the token and align it with the inputs
        cls = self.cls_quantizer(self.cls, inputs)

        # We need to broadcast the token along the batch size and set its token dimension to 1
        cls_broadcasted = tf.broadcast_to(cls, [batch_size, 1, self.hidden_size])

        return tf.concat([cls_broadcasted, inputs], 1)


@register_quantize_target(AddPositionEmbs)
@keras.saving.register_keras_serializable()
class QuantizedAddPositionEmbs(QuantizedLayer, AddPositionEmbs):
    """Quantize the :class:`AddPositionEmbs` layer, allowing operations in FixedPoint domain.

    Args:
        quant_config (dict, optional): the serialized quantization configuration. Defaults to None.
    """

    def __init__(self, *args, quant_config=None, **kwargs):
        super().__init__(*args, quant_config=quant_config, **kwargs)
        out_quant_cfg = self.quant_config.get("output_quantizer", False)
        if out_quant_cfg:
            self.out_quantizer = OutputQuantizer(
                name="output_quantizer", **out_quant_cfg)
        else:
            self.out_quantizer = None
        pe_quantizer_cfg = self.quant_config.get("pe_quantizer", {})
        self.pe_quantizer = AlignedWeightQuantizer(name="pe_quantizer", **pe_quantizer_cfg)
        self.buffer_bitwidth = apply_buffer_bitwidth(self.quant_config, signed=True)

    @tensor_inputs([QFloat])
    @rescale_outputs
    def call(self, inputs):
        # Quantize position embeddings and align them on the inputs
        pe = self.pe_quantizer(self.pe, inputs)

        # Add position embeddings
        return inputs + pe


@keras.saving.register_keras_serializable()
class ExtractToken(keras.layers.Layer):
    """ Wrapper class of `tf.gather` operation that allows to extract a Token.

    Args:
        token (int): the indice of the token to extract.
        axis (int, optional): axis over which the user gather the token. Defaults to 1.

    """

    def __init__(self, *args, token, axis=1, **kwargs):
        super().__init__(*args, **kwargs)
        self.token = token
        self.axis = axis

    def call(self, inputs):
        return tf.gather(inputs, self.token, axis=self.axis)

    def get_config(self):
        config = super().get_config()
        config.update({"token": self.token})
        config.update({"axis": self.axis})
        return config


@register_quantize_target(ExtractToken)
@register_no_output_quantizer
@register_aligned_inputs
@keras.saving.register_keras_serializable()
class QuantizedExtractToken(QuantizedLayer, ExtractToken):
    """ Quantized version of the ExtractToken layer. Accepts only FixedPoint inputs.
    """

    @tensor_inputs([FixedPoint])
    def call(self, inputs):
        return tf.gather(inputs, self.token, axis=self.axis)

    def get_config(self):
        return ExtractToken.get_config(self)



#!/usr/bin/env python
# ******************************************************************************
# Copyright 2024 Brainchip Holdings Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ******************************************************************************

__all__ = ["convert_matmul_to_gemm"]

import onnx.helper
import onnx.numpy_helper

from ..model import ONNXModel


def convert_matmul_to_gemm(model):
    """Converts Matmul node to a Gemm node.

    Args:
        model (ONNXModel): The ONNX model to be processed.
    """
    def _gemm_weight_from_matmul(matmul_node):
        # MatMul weights need to be transposed as in Gemm node we always
        # use transB=1
        matmul_weight = model.get_variable(matmul_node.input[1])
        matmul_weight = matmul_weight.T
        matmul_weight_tp = model.get_initializer(matmul_node.input[1])
        matmul_weight_tp.CopyFrom(onnx.numpy_helper.from_array(matmul_weight, matmul_node.input[1]))

    assert isinstance(model, ONNXModel)

    for node in model.nodes():
        if node.op_type == "MatMul":
            _gemm_weight_from_matmul(node)
            node.op_type = "Gemm"
            node.attribute.append(onnx.helper.make_attribute(key="transB", value=1))
    model.check_model()



#!/usr/bin/env python
# ******************************************************************************
# Copyright 2023 Brainchip Holdings Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ******************************************************************************

__all__ = ["OutputObserver"]

import tensorflow as tf
import keras

from keras.layers import Layer

from .layers_base import check_arg_constraints


@keras.saving.register_keras_serializable()
class OutputObserver(Layer):
    """ Calibration layer.

    This layer is used to compute the future `range_max` of the equivalent OutputQuantizer in the
    quantized model. It is placed where the OutputQuantizer will be inserted (end of blocks) and
    accumulates the observed maximum values (with momentum) for input in the float model.

    Args:
        axis (str): the quantization range is a scalar ('per-tensor') or a vector
            corresponding to the last axis ('per-axis'). Defaults to 'per-tensor'.
        momentum (float): the momentum for the moving average. Defaults to 0.9.
    """
    arg_constraints = {'axis': lambda: ["per-tensor", "per-axis"]}

    def __init__(self, axis="per-tensor", momentum=0.9, **kwargs):
        super().__init__(**kwargs)
        self.axis = axis
        self.momentum = momentum
        self._decay = tf.convert_to_tensor(1.0 - momentum, name="decay")
        check_arg_constraints(self, self.get_config())

    def build(self, input_shape):
        """Build the layer.

        Args:
            input_shape (list): the shape of input tensor.
        """
        super().build(input_shape)
        # Convert axis to a list of int
        if self.axis == "per-axis":
            ndims = len(input_shape)
            if ndims < 3:
                raise ValueError("OutputObserver cannot quantize per-axis tensors "
                                 " with 2 dimensions or less.")
            self._axis = list(range(len(input_shape) - 1))
        else:
            self._axis = None

        # Declares the constant/vector that will store the maximum values of the input.
        self.range_max = self.add_weight(
            name="range_max",
            shape=input_shape[-1] if self._axis is not None else (),
            dtype=tf.float32,
            initializer="ones",
            synchronization=tf.VariableSynchronization.ON_READ,
            trainable=False,
            aggregation=tf.VariableAggregation.MEAN,
            experimental_autocast=False,
        )

    def call(self, inputs):
        """ Observe inputs and update the maximum value with momentum.

        Args:
            inputs (tf.Tensor): the inputs tensor.

        Returns:
            tf.Tensor: unchanged inputs
        """
        # Compute the new range_max from inputs
        range_max = tf.math.reduce_max(tf.math.abs(inputs), self._axis)

        # If range_max was never updated set their newly computed values otherwise update with
        # moving average algorithm
        if tf.reduce_all(tf.math.equal(self.range_max, tf.constant(1.))):
            new_range_max = range_max
        else:
            # The new value is just the multiplication by decay
            old_value = self.range_max
            update_delta = (old_value - tf.cast(range_max, old_value.dtype)) * self._decay
            new_range_max = old_value - update_delta
        self.range_max.assign(new_range_max)
        return inputs

    def get_config(self):
        """Get the config of the layer.

        Returns:
            dict: the config of the layer.
        """
        config = super().get_config()
        config.update({"axis": self.axis, "momentum": self.momentum})
        return config



#!/usr/bin/env python
# ******************************************************************************
# Copyright 2024 Brainchip Holdings Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ******************************************************************************
import tensorflow as tf

from ... import layers as keras_qml_layers
from ...layers.layers_base import QuantizedLayer
from ...tensors import QTensor, QFloat, FixedPoint
from ...models.transforms import sanitize as keras_sanitize
from ..tools import generate_keras_random_samples
from .common import compare, make_fn_on_list, remove_outliers, eval_metrics, compute_saturation

skippable_layers = (keras_qml_layers.QuantizedDropout, keras_qml_layers.QuantizedFlatten,
                    keras_qml_layers.QuantizedReshape, keras_qml_layers.QuantizedRescaling)
conditional_skippable_layers = (keras_qml_layers.QuantizedMaxPool2D,
                                keras_qml_layers.QuantizedGlobalAveragePooling2D)


def _convert_to_list(x):
    if not isinstance(x, (tuple, list)):
        x = [x]
    return x


@make_fn_on_list
def _dequantize(x, /):
    return x.to_float() if isinstance(x, QTensor) else x


def _compute_mask(x, /):
    min_value = -2**x.value_bits
    max_value = 2**x.value_bits - 1
    values = x.values.numpy()
    return (values > min_value) & (values < max_value)


def _is_measurable(qlayer):
    if isinstance(qlayer, conditional_skippable_layers):
        return getattr(qlayer, "out_quantizer", None) is not None
    return not isinstance(qlayer, skippable_layers) and isinstance(qlayer, QuantizedLayer)


def _get_layer(layer_name, model):
    try:
        layer = model.get_layer(layer_name)
    except Exception:
        raise RuntimeError(f"{model.name} must have layer '{layer_name}'")
    return layer


def _search_quantized_target_layers(model, target_layer_name=None):
    if target_layer_name is not None:
        target_layers = [_get_layer(target_layer_name, model)]
    else:
        target_layers = model.layers
    # Filter target layers
    target_layers = [ly for ly in target_layers if _is_measurable(ly)]
    if len(target_layers) == 0:
        raise ValueError(f"{model.name} does not contain layers that generate quantization error!")
    return target_layers


@make_fn_on_list
def _apply_output_rescaling_rate(x, /, scales):
    assert isinstance(x, FixedPoint)
    if x.shape[-1] == tf.size(scales) or tf.size(scales) == 1:
        x = QFloat(x, scales)
    elif not tf.reduce_all(scales == 1):
        raise ValueError(f"Impossible to apply scales to {x}: incompatible shape "
                         f"({x.shape} vs {scales.shape})")
    return x


@make_fn_on_list
def _apply_input_rescaling_rate(foutput, target_layer, /, target_qlayers):
    if (out_quantizer := getattr(target_layer, 'out_quantizer', None)) is not None:
        # If inbound layer has an output quantizer apply rescaling rate to 'foutput'
        return foutput / out_quantizer.rescaling_rate
    inbound_layers = target_layer.inbound_nodes[0].inbound_layers
    if target_layer not in target_qlayers and not isinstance(inbound_layers, list):
        # Target layer is skippable, we continue to search the inbound output quantizer
        return _apply_input_rescaling_rate(foutput, inbound_layers, target_qlayers=target_qlayers)
    # There are not rescaling to apply
    return foutput


@make_fn_on_list
def compare_outputs(foutputs, qoutputs, /, per_channel=False):
    """Measures the error in a set of tensors

    Args:
        foutputs (tf.Tensor or list): the output of a float layer.
        qoutputs (QTensor or list): the quantized output to be compare with ``foutputs``.
        per_channel (bool, optional): comparison is done for each channel. Defaults to False.

    Returns:
        dict or list: the quantization error.
    """
    assert isinstance(qoutputs, QTensor), f"{qoutputs} must be a QTensor"
    axis = -1 if per_channel else None

    # Compute saturation and mask where is indicated if a value saturate or not
    saturation = compute_saturation(qoutputs, axis=axis)
    mask = _compute_mask(qoutputs)

    # Dequantize and convert tensors to numpy
    foutputs = foutputs.numpy()
    qoutputs = qoutputs.to_float().numpy()

    # Exclude samples that saturate since the error is ambiguous out of range.
    foutputs = remove_outliers(foutputs, mask, axis=axis)
    qoutputs = remove_outliers(qoutputs, mask, axis=axis)
    return compare(foutputs, qoutputs, saturation)


def quantization_error(fmodel, qmodel, target_layer=None, batch_size=1, seed=None):
    """Measures the layer quantization error in a set of Keras models

    Args:
        fmodel (tf.keras.Model): the float model.
        qmodel (tf.keras.Model): the quantized version of `fmodel`.
        target_layer (str, optional): computation error is performed only in the target layer,
            expanding the analysis to each output channel. Defaults to None.
        batch_size (int, optional): the batch size. Defaults to 1.
        seed (int, optional): a random seed. Defaults to None.

    Returns:
        dict: the quantization error of the target layers
    """
    per_channel = target_layer is not None

    # Sanitize float model
    fmodel = keras_sanitize(fmodel)

    # Create an intermediary quantized model that will compute the inputs for both models
    target_qlayers = _search_quantized_target_layers(qmodel, target_layer_name=target_layer)
    qmodel = tf.keras.Model(qmodel.input, [qly.input for qly in target_qlayers])

    # Generate a random set of samples
    samples = generate_keras_random_samples(qmodel, batch_size=batch_size, seed=seed)

    # Compute quantization error per layer:
    # Generate the set of input quantized samples
    summary = {}
    qinputs = qmodel(samples)
    for qlayer, qx in zip(target_qlayers, _convert_to_list(qinputs)):
        # Match quantized layer in fmodel
        flayer = _get_layer(qlayer.name, fmodel)

        # Forward qx in both qlayer and flayer.
        # Note there is no error provided by inputs quantization
        qoutputs = qlayer(qx)

        # Since qinputs will be used as inputs of flayer, they potentially have a
        # rescaling_rate factor. Since we cannot remove this factor from qinputs because qlayer
        # expects it (CLE modified its weights in favor of), we are forced to scale finputs
        # to produce consistent inputs in both layers.
        # Note this will happend per each input (in case of multi-inbounds)
        finputs = _apply_input_rescaling_rate(_dequantize(qx),
                                              qlayer.inbound_nodes[0].inbound_layers,
                                              target_qlayers=target_qlayers)
        foutputs = flayer(finputs)

        # Due to the CLE, the outputs of quantized layers with output quantizer are not comparable
        # to their float version, since rescaling_rate is only applied to the weights of the
        # next layer. Therefore, quantized outputs must be rescaled to be equivalent.
        if (out_quantizer := getattr(qlayer, "out_quantizer", None)):
            scale_factor = 1 / out_quantizer.rescaling_rate
            qoutputs = _apply_output_rescaling_rate(qoutputs, scales=scale_factor)

        # Compute quantization error per layer
        key = f"{qlayer.name} ({qlayer.__class__.__name__})"
        summary[key] = eval_metrics(compare_outputs(foutputs, qoutputs, per_channel=per_channel))
    return summary


def cumulative_quantization_error(fmodel, qmodel, target_layer=None, batch_size=1, seed=None):
    """Measures the cumulative quantization error in a set of Keras models

    Args:
        fmodel (tf.keras.Model): the float model.
        qmodel (tf.keras.Model): the quantized version of `fmodel`.
        target_layer (str, optional): error computation is performed only in the target layer,
            expanding the analysis to each output channel. Defaults to None.
        batch_size (int, optional): the batch size. Defaults to 1.
        seed (int, optional): a random seed. Defaults to None.

    Returns:
        dict: the quantization error by each layer
    """
    per_channel = target_layer is not None

    # Sanitize float model
    fmodel = keras_sanitize(fmodel)

    # Create intermediary models with the target outputs
    target_qlayers = _search_quantized_target_layers(qmodel, target_layer_name=target_layer)
    qmodel = tf.keras.Model(qmodel.input, [qly.output for qly in target_qlayers])
    foutputs = [_get_layer(ly.name, fmodel).output for ly in target_qlayers]
    fmodel = tf.keras.Model(fmodel.input, foutputs)

    # Generate a random set of samples
    samples = generate_keras_random_samples(qmodel, batch_size=batch_size, seed=seed)

    # Compute cumulative quantization error
    summary = {}
    outputs = _convert_to_list(fmodel(samples)), _convert_to_list(qmodel(samples))
    for qlayer, foutputs, qoutputs in zip(target_qlayers, *outputs):
        key = f"{qlayer.name} ({qlayer.__class__.__name__})"

        # Due to the CLE, the outputs of quantized layers with output quantizer are not
        # comparable to their float version, since rescaling_rate is only applied to
        # the weights of the next layer. Therefore, quantized outputs must be rescaled
        # to be equivalent.
        if (out_quantizer := getattr(qlayer, "out_quantizer", None)) is not None:
            scale_factor = 1 / out_quantizer.rescaling_rate
            qoutputs = _apply_output_rescaling_rate(qoutputs, scales=scale_factor)

        # Compute cumulative quantization error
        summary[key] = eval_metrics(compare_outputs(foutputs, qoutputs, per_channel=per_channel))
    return summary



#!/usr/bin/env python
# ******************************************************************************
# Copyright 2023 Brainchip Holdings Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ******************************************************************************
import numpy as np
from onnxruntime.quantization.calibrate import TensorData

from ..graph_tools import get_tensor_shape, get_field, find_value_info_by_name

# Akida inputs/outputs range is [-127, 127], int8 for hidden layers
AKIDA_IO_MAX = 127.0


def needs_zp(graph, tensor_name):
    tensor_value_info = find_value_info_by_name(graph, tensor_name)
    tensor_shape = get_tensor_shape(tensor_value_info)

    # Search nodes link to tensor_name
    nodes_link_to_tensor = []
    for node in graph.node:
        if tensor_name in node.input:
            nodes_link_to_tensor.append(node)

    # Support zero point if all nodes meet the condition
    allow_zp = True
    for input_node in nodes_link_to_tensor:
        # Retrieve node properties
        op_type = input_node.op_type
        group = get_field(input_node, "group", 1)

        # Model may have zero point if input op is conv and channel numbers is one of [1, 3].
        # Note: shape format is (batch,C,X,Y).
        allow_zp = allow_zp and op_type == 'Conv' and tensor_shape[1] in (1, 3) and group == 1
    return allow_zp


def _input_conv_zp_scale(input_range):
    assert isinstance(input_range, TensorData)
    rmin, rmax = input_range.range_value
    if np.any(rmin >= rmax):
        raise ValueError("Invalid input range")
    # input is uint8, so max is 255. Hence we can deduce the scale
    # Note that akida_scale is reciprocal of onnx scale
    akida_scale = 255 / (rmax - rmin)
    zero_point = -np.round(rmin * akida_scale)
    # In hardware we cannot handle negative zero point. But a negative zero point is
    # a consequence of an input WITH POSITIVE RANGE. For this case, we can quantize assuming
    # a symmetric range between [-rmax, rmax] (rmin = 0).
    akida_scale = np.where(zero_point < 0, 255 / rmax, akida_scale)
    zero_point = np.maximum(0, zero_point)
    return akida_scale, np.array(zero_point, np.uint8)


def input_scale_no_zp(input_range):
    assert isinstance(input_range, TensorData)
    rmin, rmax = input_range.range_value
    if np.any(rmin > rmax):
        raise ValueError("Invalid input range")
    rmax = np.maximum(np.abs(rmin), np.abs(rmax))
    # Replace rmax == 0 by an epsilon to avoid division by zero
    rmax = np.maximum(rmax, 1e-7)
    # input is int8, so max is AKIDA_IO_MAX. Hence we can deduce the scale
    # Note that akida_scale is reciprocal of onnx scale
    akida_scale = AKIDA_IO_MAX / rmax
    return akida_scale


def input_zp_scale(input_range, allow_zp=False):
    """Compute the input scale and zero point """
    if allow_zp:
        i_scale, zero_point = _input_conv_zp_scale(input_range)
    else:
        # this will be like an input data + conv, no zero point
        # Note: To force signed QuantizeLinear outputs, we return an int8 zero point
        i_scale = input_scale_no_zp(input_range)
        zero_point = np.zeros_like(i_scale, dtype=np.int8)
    return np.array(i_scale, "float64"), zero_point
